"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

See the Licence for the specific language governing permissions and limitations under the Licence.

"""
from __future__ import (absolute_import, print_function, unicode_literals)

from future.backports import OrderedDict
from future.utils import with_metaclass
from nine import (iteritems, str, range, map, nine)

import warnings
import copy
import getopt
import sys
import datetime
import os
import pprint
import inspect
from collections import namedtuple
import multiprocessing
import cftime
import threading
import xml.dom.minidom
import pcraster
from netCDF4 import Dataset, date2num, num2date
from pandas.core.tools.datetimes import parsing
import numpy as np

from .errors import LisfloodError, LisfloodWarning, LisfloodFileError
from .decorators import cached
from .default_options import default_options

project_dir = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '../../..'))


class Singleton(type):
    """
    Singleton metaclass to keep single instances by init arguments
    """
    _instances = {}
    _init = {}
    _current = {}

    def __init__(cls, name, bases, dct):
        cls._init[cls] = dct.get('__init__', None)  # set __init__ method for the class
        super(Singleton, cls).__init__(name, bases, dct)

    def __call__(cls, *args, **kwargs):
        init = cls._init[cls]  # get __init__ method
        if init is not None:
            init_args = inspect.getcallargs(init, None, *args, **kwargs).items()
            new_init_args = []
            for a in init_args:
                if isinstance(a[1], np.ndarray):
                    new_init_args.append((a[0], a[1].tostring()))
                else:
                    new_init_args.append(a)
            key = (cls, frozenset(new_init_args))
        else:
            key = cls

        if key not in cls._instances:
            cls._instances[key] = super(Singleton, cls).__call__(*args, **kwargs)
        cls._current[cls] = cls._instances[key]
        return cls._instances[key]

    def instance(cls):
        return cls._current[cls]


class ThreadSingleton(type):
    """
    Singleton metaclass to keep single instances by init arguments and by thread
    """
    _instances = {}
    _current = {}

    def __init__(cls, name, bases, dct):
        super(ThreadSingleton, cls).__init__(name, bases, dct)

    def __call__(cls, *args, **kwargs):
        init = cls.__init__
        task_id = cls.get_task_id()
        if init is not None:
            init_args = inspect.getcallargs(init, None, *args, **kwargs).items()
            new_init_args = []
            for a in init_args:
                if isinstance(a[1], np.ndarray):
                    new_init_args.append((a[0], a[1].tostring()))
                else:
                    new_init_args.append(a)
            key = (task_id, cls, str(new_init_args))
        else:
            key = (task_id, cls)

        if key not in cls._instances:
            cls._instances[key] = super(ThreadSingleton, cls).__call__(*args, **kwargs)
        cls._current[(task_id, cls)] = cls._instances[key]
        return cls._instances[key]

    def instance(cls):
        task_id = cls.get_task_id()
        key = (task_id, cls)
        return cls._current[key]

    def get_task_id(cls):
        task_id = multiprocessing.current_process().name + '_' + str(threading.get_ident())
        return task_id

@nine
class CDFFlags(with_metaclass(Singleton)):
    """
    Flags for netcdf output for end, steps, all, monthly (steps), yearly (steps), monthly , yearly
    flagcdf = 0
    flagcdf = 1  # index flag for writing nedcdf = 1 (=steps) -> indicated if a netcdf is created or maps are appended
    flagcdf = 2
    flagcdf = 3  # set to monthly (step) flag
    flagcdf = 4  # set to yearly (step) flag
    flagcdf = 5  # set to monthly flag
    flagcdf = 6  # set to yearly flag
    """
    def __init__(self, uid):
        self.uid = uid
        # FIXME magic numbers. replace indexes with descriptive keys (end, steps, all, monthly (steps), etc.)
        self.flags = [0, 0, 0, 0, 0, 0, 0]

    def inc(self, idx):
        self.flags[idx] += 1

    def set(self, data):
        self.flags = data

    def __getitem__(self, item):
        return self.flags[item]


@nine
class MaskAttrs(with_metaclass(Singleton)):
    def __init__(self, maskname):
        self.name = maskname
        self._attrs = {
            'x': pcraster.clone().west(),
            'y': pcraster.clone().north(),
            'col': pcraster.clone().nrCols(),
            'row': pcraster.clone().nrRows(),
            'cell': pcraster.clone().cellSize(),
        }

    def __getitem__(self, item):
        return self._attrs.get(item)


@nine
class CutMap(with_metaclass(Singleton)):
    def __init__(self, cut1, cut2, cut3, cut4):
        self.cuts = (cut1, cut2, cut3, cut4)


@nine
class MaskInfo(with_metaclass(Singleton)):
    Info = namedtuple('Info', 'mask, shape, maskflat, shapeflat, mapC, maskall')

    def __init__(self, mask, maskmap):
        self.flat = mask.ravel()
        self.mask_compressed = np.ma.compressed(np.ma.masked_array(mask, mask))  # mapC
        self.mask_all = np.ma.masked_all(self.flat.shape)
        self.mask_all.mask = self.flat
        self.info = self.Info(mask, mask.shape, self.flat, self.flat.shape, self.mask_compressed.shape, self.mask_all)
        self._in_zero = np.zeros(self.mask_compressed.shape)
        self.maskmap = maskmap

    def in_zero(self):
        return self._in_zero.copy()

    def __iter__(self):
        return iter(self.info)


@nine
class NetCDFMetadata(with_metaclass(Singleton)):
    def __init__(self, uid):
        from .zusatz import iterOpenNetcdf
        self.uid = uid
        self.data = {}
        settings = LisSettings.instance()
        binding = settings.binding
        try:
            # using ncdftemplate
            filename = os.path.splitext(binding['netCDFtemplate'])[0] + '.nc'
            nf1 = iterOpenNetcdf(filename, "Trying to get metadata from netcdf template \n", 'r')
            for var in nf1.variables:
                self.data[var] = {k: v for k, v in iteritems(nf1.variables[var].__dict__) if k != '_FillValue'}
            nf1.close()
            return
        except (KeyError, IOError, IndexError, Exception):
            pass
        # if no template .nc is given the e.nc file is used
        filename = os.path.splitext(binding['E0Maps'])[0] + '.nc'
        nf1 = iterOpenNetcdf(filename, "Trying to get metadata from E0 maps \n", 'r')
        for var in nf1.variables:
            self.data[var] = {k: v for k, v in iteritems(nf1.variables[var].__dict__) if k != '_FillValue'}
        nf1.close()

@nine
class EPICSettings(with_metaclass(Singleton)):

    def __init__(self):
        self.soil_uses = ["Rainfed", 'Forest', 'Irrigated']
        self.prescribed_vegetation = ['{}_prescribed'.format(fract) for fract in self.soil_uses]
        self.vegetation_landuse = OrderedDict(
            zip(self.prescribed_vegetation[:], self.soil_uses[:])
        )
        self.landuse_vegetation = OrderedDict(
            [(v, [k]) for k, v in self.vegetation_landuse.items()]
        )
        self.prescribe_lai = OrderedDict(
            zip(self.prescribed_vegetation[:], ['LAIOtherMaps', 'LAIForestMaps', 'LAIIrrigationMaps'])
        )
        self.landuse_inputmap = OrderedDict(
                zip(self.landuse_vegetation.keys(), ["OtherFraction", "ForestFraction", "IrrigationFraction"])
        )


@nine
class LisSettings(with_metaclass(ThreadSingleton)):
    printer = pprint.PrettyPrinter(indent=4, width=120)

    def __str__(self):
        res = """
    Binding: {binding}
    Options: {options}
    report_steps: {report_steps}
    report_timeseries: {report_timeseries}
    report_maps_steps: {report_maps_steps}
    report_maps_all: {report_maps_all}
    report_maps_end: {report_maps_end}
    """.format(binding=self.printer.pformat(self.binding), options=self.printer.pformat(self.options),
               report_steps=self.printer.pformat(self.report_steps),
               report_timeseries=self.printer.pformat(self.report_timeseries),
               report_maps_steps=self.printer.pformat(self.report_maps_steps),
               report_maps_all=self.printer.pformat(self.report_maps_all),
               report_maps_end=self.printer.pformat(self.report_maps_end))
        return res

    def __init__(self, settings_file, sys_args=[]):
        dom = xml.dom.minidom.parse(settings_file)
        self.settings_dir = os.path.normpath(os.path.dirname((os.path.abspath(settings_file))))
        self.settings_path = os.path.normpath(os.path.abspath(settings_file))
        user_settings, bindings = self._bindings(dom)
        self.timestep_init = None if not bindings.get('timestepInit') else bindings['timestepInit']
        self._check_timestep_init()
        self.output_dir = self._out_dir(user_settings)
        self.ncores = self._ncores(user_settings)
        self.binding = bindings
        self.options = self._options(dom)
        self.flags = self._flags(sys_args)
        self.model_steps = self._model_steps()
        self.report_steps = self._report_steps(user_settings, bindings)
        self.filter_steps = self._filter_steps(user_settings)
        self.ens_members = self._ens_members(user_settings)
        self.report_timeseries, self.report_maps_steps, self.report_maps_all, self.report_maps_end = self.build_reportedmaps_dicts()
        self.enkf_set, self.mc_set = self.montecarlo_kalman_settings()
        self.step_start, self.step_end = self.binding['StepStart'], self.binding['StepEnd']
        self.step_start_int, self.step_end_int = self.binding['StepStartInt'], self.binding['StepEndInt']
        ref_date_start = calendar(self.binding['CalendarDayStart'], self.binding['calendar_type'])
        self.step_start_dt = inttodate(self.step_start_int - 1, ref_date_start, binding=self.binding)
        self.step_end_dt = inttodate(self.step_end_int - 1, ref_date_start, binding=self.binding)
        self.maskpath = self.binding['MaskMap']

    def build_reportedmaps_dicts(self):
        self.report_timeseries = self._report_tss()
        self.report_maps_steps, self.report_maps_all, self.report_maps_end = self._reported_maps()
        self.report_maps_steps = {k: v for k, v in iteritems(self.report_maps_steps) if v}
        self.report_maps_all = {k: v for k, v in iteritems(self.report_maps_all) if v}
        self.report_maps_end = {k: v for k, v in iteritems(self.report_maps_end) if v}
        return self.report_timeseries, self.report_maps_steps, self.report_maps_all, self.report_maps_end

    def montecarlo_kalman_settings(self):
        # Ensemble Kalman filter
        enkf_set = self.options.get('EnKF', 0) if not self.options['InitLisflood'] else 0
        # MonteCarlo
        mc_set = self.options.get('MonteCarlo', 0) if not self.options['InitLisflood'] else 0
        if enkf_set and not mc_set:
            msg = "Trying to run EnKF with only 1 ensemble member \n"
            raise LisfloodError(msg)
        if enkf_set and self.filter_steps[0] == 0:
            msg = "Trying to run EnKF without filter timestep specified \nRunning LISFLOOD in Monte Carlo mode \n"
            warnings.warn(LisfloodWarning(msg))
            enkf_set = 0
        if mc_set and self.ens_members[0] <= 1:
            msg = "Trying to run Monte Carlo simulation with only 1 member \nRunning LISFLOOD in deterministic mode \n"
            warnings.warn(LisfloodWarning(msg))
            mc_set = 0
        return enkf_set, mc_set

    def _check_timestep_init(self):
        try:
            float(self.timestep_init)
        except ValueError:
            try:
                parsing.parse_time_string(self.timestep_init, dayfirst=True)
            except ValueError:
                raise LisfloodError('Option timestepInit was not parsable. Must be integer or date string: {}'.format(self.timestep_init))
            else:
                return True
        else:
            return True

    def _check_simulation_dates(self):
        """ Check simulation start and end dates or timesteps

        Check simulation start and end dates/timesteps to be later than begin date (CalendarStartDay).
        If dates are used for binding[start] and binding[end], it substitutes dates with time step numbers.

        :return int_start, int_end
        :rtype tuple
        :raise LisfloodError if dates are not compatible
        """
        begin = calendar(self.binding['CalendarDayStart'], self.binding['calendar_type'])

        int_start, str_start = datetoint(self.binding['StepStart'], self.binding)
        int_end, str_end = datetoint(self.binding['StepEnd'], self.binding)

        # test if start and end > begin
        if (int_start < 0) or (int_end < 0) or ((int_end - int_start) < 0):
            str_begin = begin.strftime("%d/%m/%Y %H:%M")
            msg = "Simulation start date and/or simulation end date are wrong or do not match CalendarStartDate!\n" + \
                  "CalendarStartDay: " + str_begin + "\n" + \
                  "Simulation start: " + str_start + " - " + str(int_start) + "\n" + \
                  "Simulation end: " + str_end + " - " + str(int_end)
            raise LisfloodError(msg)
        self.binding['StepStartInt'] = int_start
        self.binding['StepEndInt'] = int_end
        return int_start, int_end

    def _model_steps(self):
        int_start, int_end = self._check_simulation_dates()
        return [int_start, int_end]

    @staticmethod
    def _ncores(user_settings):
        # Number of cores for MonteCarlo or EnKF
        nr_cores = []
        try:
            nr_cores.append(int(user_settings['nrCores']))
        except (KeyError, ValueError):
            nr_cores.append(1)
        return nr_cores

    @staticmethod
    def _ens_members(user_settings):
        # Number of Ensemble members for MonteCarlo or EnKF
        res = []
        try:
            res.append(int(user_settings["EnsMembers"]))
        except (KeyError, ValueError):
            res.append(1)
        return res

    @staticmethod
    def _out_dir(user_settings):
        # if pathout has some placeholders, they are replaced here
        pathout = user_settings["PathOut"]
        while pathout.find('$(') > -1:
            a1 = pathout.find('$(')
            a2 = pathout.find(')')
            try:
                s2 = user_settings[pathout[a1 + 2:a2]]
            except KeyError:
                print('no ', pathout[a1 + 2:a2], 'for', pathout, ' in lfuser defined')
            else:
                pathout = pathout.replace(pathout[a1:a2 + 1], s2)

        # CM: output folder
        return pathout

    @staticmethod
    def _flags(sys_args):
        flags = OrderedDict([('quiet', False), ('veryquiet', False), ('loud', False),
                             ('checkfiles', False), ('noheader', False), ('printtime', False),
                             ('debug', False), ('nancheck', False), ('initonly', False)])

        @cached
        def _flags(argz):
            try:
                opts, arguments = getopt.getopt(argz, 'qvlchtdni', list(flags.keys()))
            except getopt.GetoptError:
                from ..main import usage
                usage()
            else:
                for o, a in opts:
                    for opt in (('-q', '--quiet'), ('-v', '--veryquiet'),
                                ('-l', '--loud'), ('-c', '--checkfiles'),
                                ('-h', '--noheader'), ('-t', '--printtime'),
                                ('-d', '--debug'), ('-n', '--nancheck'),
                                ('-i', '--initonly')):
                        if o in opt:
                            flags[opt[1].lstrip('--')] = True
                            break
            return flags

        return _flags(sys_args)

    def _bindings(self, dom):
        binding = {}

        #  built-in user variables
        user = {
            'ProjectDir': project_dir, 'ProjectPath': project_dir,
            'SettingsDir': self.settings_dir, 'SettingsPath': self.settings_dir,
        }
        # get all the bindings in the first part of the settingsfile = lfuser
        # list of elements "lfuser" in settings file
        lfuse = dom.getElementsByTagName("lfuser")[0]
        for userset in lfuse.getElementsByTagName("textvar"):
            user[userset.attributes['name'].value] = str(userset.attributes['value'].value)
        # get all the binding in the last part of the settingsfile  = lfbinding
        # list of elements "lfbinding" in settings file
        bind = dom.getElementsByTagName("lfbinding")[0]
        for bindset in bind.getElementsByTagName("textvar"):
            binding[bindset.attributes['name'].value] = str(bindset.attributes['value'].value)

        # replace/add the information from lfuser to lfbinding
        for i in binding:
            expr = binding[i]
            while expr.find('$(') > -1:
                a1 = expr.find('$(')
                a2 = expr.find(')')
                try:
                    s2 = user[expr[a1 + 2:a2]]
                except KeyError:
                    print('no ', expr[a1 + 2:a2], 'for', binding[i], ' in lfuser defined')
                expr = expr.replace(expr[a1:a2 + 1], s2)
            binding[i] = expr

        # set the calendar type
        binding["calendar_type"] = binding['CalendarConvention']

        return user, binding

    @staticmethod
    def _report_steps(user_settings, bindings):

        res = {}
        repsteps = user_settings['ReportSteps'].split(',')
        if repsteps[0] == 'starttime':
            repsteps[0] = str(bindings['StepStartInt'])
        if repsteps[-1] == 'endtime':
            repsteps[-1] = str(bindings['StepEndInt'])
        jjj = []
        for i in repsteps:
            if '..' in i:
                j = list(map(int, i.split('..')))
                jjj = list(range(j[0], j[1] + 1))
            else:
                jjj.append(i)
        res['rep'] = list(map(int, jjj))
        if res['rep'][0] > bindings['StepEndInt'] or res['rep'][-1] < bindings['StepStartInt']:
            warnings.warn(LisfloodWarning('No maps are reported as report steps configuration is outside simulation time interval'))
        return res

    @staticmethod
    def _options(settingsfile_dom):
        options = copy.deepcopy(default_options)
        # getting option set in the specific settings file
        # and resetting them to their choice value
        lfoptions_elem = settingsfile_dom.getElementsByTagName("lfoptions")[0]
        option_setting = {}
        for optset in lfoptions_elem.getElementsByTagName("setoption"):
            option_setting[optset.attributes['name'].value] = bool(int(optset.attributes['choice'].value))
        # overwrite defaults
        options.update(option_setting)
        options['nonInit'] = not options['InitLisflood']
        return options

    def _filter_steps(self, user_settings):
        # Split the string FilterSteps into an int array
        # remove endtime if present
        # replace .. with sequence
        res = []
        try:
            filter_steps = user_settings['FilterSteps'].split(',')
        except KeyError:
            filter_steps = 0

        if filter_steps[-1] in ('endtime', self.binding['StepEnd']):
            filter_steps[-1] = 0

        for i in filter_steps:
            try:
                time_dif = datetime.datetime.strptime(i, "%d/%m/%Y") - datetime.datetime.strptime(self.binding['CalendarDayStart'], "%d/%m/%Y")
                val = time_dif.days
            except:
                val = int(i)

            stependint = self.binding['StepEndInt']
            # if int(val) < int(binding['StepEnd']):
            if val < stependint:
                try:
                    res.append(int(i))
                except:
                    res.append(time_dif.days)
        return res

    def _report_tss(self):
        report_time_series_act = {}
        # running through all times series
        timeseries = self.options['timeseries']
        for ts in timeseries.values():
            ts_obj = self._set_active_options(ts, ts.repoption, ts.restrictoption)
            if not ts_obj:
                continue
            key = ts.name
            report_time_series_act[key] = self._set_active_options(ts, ts.repoption, ts.restrictoption)

        return report_time_series_act

    def _reported_maps(self):
        report_maps_steps = {}
        report_maps_all = {}
        report_maps_end = {}

        # running through all maps
        reportedmaps = self.options['reportedmaps']
        for rm in reportedmaps.values():
            key = rm.name
            report_maps_all[key] = self._set_active_options(rm, rm.all, rm.restrictoption)
            report_maps_steps[key] = self._set_active_options(rm, rm.steps, rm.restrictoption)
            report_maps_end[key] = self._set_active_options(rm, rm.end, rm.restrictoption)

        return report_maps_steps, report_maps_all, report_maps_end

    def _set_active_options(self, obj, report_options, restricted_options):
        """
        Check report options (when to report/write the map/tss: at least one true)
        and restricted_options (all options must be enabled)
        Return ReportMap tuple if is to write, according report and restrict options
        :param obj: `lisflood.global_modules.default_options.ReportedMap/TimeSeries` tuple
        :param report_options: list of option keys enabling report
        :param restricted_options: list of option keys disabling report (has precedence over report_options)
        :return: obj or None
        """
        allow = any(self.options.get(repopt) for repopt in report_options)
        # checking that at least one restricted_options is set
        if allow and restricted_options:
            allow = all(self.options.get(ro) for ro in restricted_options)
        return obj if allow else None


def get_calendar_type(nc):
    """Get the type of calendar of the open netCDF file passed as argument (http://cfconventions.org/)"""
    try:
        calendar_type = nc.variables["time"].calendar
    except AttributeError:
        calendar_type = "proleptic_gregorian"
        warnings.warn(LisfloodWarning("""
The 'calendar' attribute of the 'time' variable of {} is not set: the default '{}' is used
(http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.pdf)
""".format(nc, calendar_type)))
    return calendar_type


def calendar_inconsistency_warning(filename, file_calendar, precipitation_calendar):
    return LisfloodWarning("In file {}, time.calendar attribute ({}) differs from that of precipitation ({})!".format(filename, file_calendar, precipitation_calendar))


def calendar(date_in, calendar_type='proleptic_gregorian'):
    """ Get date or number of steps from input.

    Get date from input string using one of the available formats or get time step number from input number or string.
    Used to get the date from CalendarDayStart (input) in the settings xml

    :param date_in: string containing a date in one of the available formats or time step number as number or string
    :param calendar_type:
    :rtype: datetime object or float number
    :returns: date as datetime or time step number as float
    :raises ValueError: stop if input is not a step number AND it is in wrong date format
    """
    try:
        # try reading step number from number or string
        return float(date_in)
    except ValueError:
        # try reading a date in one of available formats
        try:
            _t_units = "hours since 1970-01-01 00:00:00"  # units used for date type conversion (datetime.datetime -> calendar-specific if needed)
            date = parsing.parse_time_string(date_in, dayfirst=True)[0]  # datetime.datetime type
            step = date2num(date, _t_units, calendar_type)  # float type
            return num2date(step, _t_units, calendar_type)  # calendar-dependent type from netCDF4.netcdftime._netcdftime module
        except:
            # if cannot read input then stop
            msg = "Wrong step or date format in XML settings file\n Input {}".format(date_in)
            raise LisfloodError(msg)


def datetoint(date_in, binding=None):
    """ Get number of steps between dateIn and CalendarDayStart.

    Get the number of steps between dateIn and CalendarDayStart and return it as integer number.
    It can now compute the number of sub-daily steps.
    dateIn can be either a date or a number. If dateIn is a number, it must be the number of steps between
    dateIn and CalendarDayStart.

    :param date_in: date as string or number
    :param binding: obptional binding dictionary
    :return number of steps as integer and input date as string
    :rtype tuple(<int>, <str>)
    """
    if not binding:
        settings = LisSettings.instance()
        binding = settings.binding
    # get reference date to be used with step numbers from 'CalendarDayStart' in Settings.xml file
    date1 = calendar(date_in, binding['calendar_type'])
    begin = calendar(binding['CalendarDayStart'], binding['calendar_type'])
    # get model time step as float form 'DtSec' in Settings.xml file
    dt_sec = float(binding['DtSec'])
    # compute fraction of day corresponding to model time step as float
    # DtDay = float(DtSec / 86400.)
    # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)

    if isinstance(date1, (datetime.datetime, cftime.DatetimeProlepticGregorian, cftime.real_datetime)):
        str1 = date1.strftime("%d/%m/%Y %H:%M")
        # get total number of seconds corresponding to the time interval between dateIn and CalendarDayStart
        timeinterval_in_sec = int((date1 - begin).total_seconds())
        # get total number of steps between dateIn and CalendarDayStart
        int1 = int(timeinterval_in_sec/dt_sec + 1)
        # int1 = (date1 - begin).days + 1
    else:
        int1 = int(date1)
        str1 = str(date1)
    return int1, str1


def inttodate(int_in, ref_date, binding=None):
    """ Get date corresponding to a number of steps from a reference date.

    Get date corresponding to a number of steps from a reference date and return it as datetime.
    It can now use sub-daily steps.
    intIn is a number of steps from the reference date refDate.

    :param int_in: number of steps as integer
    :param ref_date: reference date as datetime
    :return: stepDate: date as datetime corresponding to intIn steps from refDate
    """
    if not binding:
        settings = LisSettings.instance()
        binding = settings.binding

    # CM: get model time step as float form 'DtSec' in Settings.xml file
    DtSec = float(binding['DtSec'])
    # CM: compute fraction of day corresponding to model time step as float
    DtDay = DtSec / 86400.
    # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)

    # CM: compute date corresponding to intIn steps from reference date refDate
    stepDate = ref_date + datetime.timedelta(days=(int_in * DtDay))

    return stepDate


class LisfloodRunInfo(Warning):
    """
    the error handling class
    prints out an error
    """
    modes = {
        'MonteCarloFramework': 'Monte Carlo',
        'EnsKalmanFilterFramework': '"Ensemble Kalman Filter',
        'DynamicFramework': 'Deterministic',
    }

    def __init__(self, model):

        header = "\n\n ========================== LISFLOOD Simulation Information and Setting =============================\n"
        msg = ''
        settings = LisSettings.instance()
        option = settings.options
        out_dir = settings.output_dir
        ens_members = settings.ens_members[0]
        nr_cores = settings.ncores[0]
        steps = len(settings.filter_steps)
        mode = self.modes[model.__class__.__name__]
        activated_options = ','.join(k for k, v in option.items() if v and not k.startswith('rep') and 'netcdf' not in k.lower() and k not in ('timeseries',))
        activated_rep = ','.join(k for k, v in option.items() if v and k.startswith('rep') and k not in 'timeseries')
        msg += "\t[X] LISFLOOD is used in the {}\n".format(mode)
        if option['InitLisflood']:
            msg += "\t[X] INITIALISATION RUN\n"

        if mode in (self.modes['EnsKalmanFilterFramework'], self.modes['MonteCarloFramework']) and ens_members > 1:
            msg += "\t[X] It uses {} ensemble members for the simulation\n".format(str(ens_members))

        if mode in (self.modes['EnsKalmanFilterFramework']) and steps > 1:
            msg += "\t[X] The model will be updated at {} time step during the simulation\n".format(str(steps))

        if mode in (self.modes['EnsKalmanFilterFramework'], self.modes['MonteCarloFramework']) and nr_cores > 1:
            msg += "\t[X] The simulation will try to use {} processors simultaneous\n".format(str(nr_cores))
        msg += "\t[X] The simulation output as specified in the settings file can be found in {}\n".format(out_dir)
        msg += "\t[X] Activated modules: {}\n".format(activated_options)
        msg += "\t[X] Report options: {}\n".format(activated_rep)
        self._msg = '{}{}'.format(header, msg)

    def __str__(self):
        return self._msg



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""
from __future__ import print_function, absolute_import

import os
import uuid
from collections import OrderedDict

import numpy as np
try:
    from numba import njit, set_num_threads, get_num_threads, config as numba_config
    numba_full = True
except ImportError:
    import numba
    print('WARNING! Numba version is probably too old. Only limited functionalities will be available for Numba')
    print('Numba version: {}'.format(numba.__version__))
    from numba import njit, config as numba_config
    numba_full = False


from .global_modules.settings import CutMap, LisSettings, NetCDFMetadata, EPICSettings, MaskInfo
from .global_modules.zusatz import DynamicModel
from .global_modules.add1 import loadsetclone, mapattrNetCDF
from .hydrological_modules.miscInitial import miscInitial

from .hydrological_modules.readmeteo import readmeteo
from .hydrological_modules.leafarea import leafarea
from .hydrological_modules.landusechange import landusechange
from .hydrological_modules.snow import snow
from .hydrological_modules.inflow import inflow
from .hydrological_modules.frost import frost
from .hydrological_modules.soil import soil
from .hydrological_modules.routing import routing
from .hydrological_modules.groundwater import groundwater
from .hydrological_modules.surface_routing import surface_routing
from .hydrological_modules.reservoir import reservoir
from .hydrological_modules.lakes import lakes
from .hydrological_modules.polder import polder
from .hydrological_modules.waterabstraction import waterabstraction
from .hydrological_modules.indicatorcalc import indicatorcalc
from .hydrological_modules.riceirrigation import riceirrigation
from .hydrological_modules.evapowater import evapowater
from .hydrological_modules.transmission import transmission
from .hydrological_modules.soilloop import soilloop
from .hydrological_modules.opensealed import opensealed
from .hydrological_modules.waterbalance import waterbalance
from .hydrological_modules.waterlevel import waterlevel
from .hydrological_modules.structures import structures

from .global_modules.output import outputTssMap
from .global_modules.stateVar import stateVar
from .global_modules.add1 import readInputWithBackup
from .global_modules.add1 import NumpyModified

@njit(fastmath=False, cache=True)
def _vegSum(ax_veg, variable, soil_fracs):
    return (soil_fracs * variable).sum(ax_veg)


# --------------------------------------------
class LisfloodModel_ini(DynamicModel):

    """ LISFLOOD initial part
        same as the PCRaster script -initial-
        this part is to initialize the variables
        it will call the initial part of the hydrological modules
    """

    def __init__(self):
        """ init part of the initial part
            defines the mask map and the outlet points
            initialization of the hydrological modules
        """
        DynamicModel.__init__(self)

        # try to make the maskmap more flexible e.g. col, row,x1,y1  or x1,x2,y1,y2
        self.MaskMap = loadsetclone('MaskMap')
        self.epic_settings = EPICSettings()
        self.settings = LisSettings.instance()
        self.maskinfo = MaskInfo.instance()
        binding = self.settings.binding
        option = self.settings.options
        flags = self.settings.flags
        report_steps = self.settings.report_steps

        # set the maximum number of threads that numba should use (now used in soilloop only)
        num_threads = int(binding["numCPUs_parallelNumba"])
        if (num_threads>0):
            if num_threads<=numba_config.NUMBA_NUM_THREADS:
                set_num_threads(num_threads)

        # Mapping of vegetation types to land use fractions (and the other way around)
        ##global VEGETATION_LANDUSE, LANDUSE_VEGETATION, PRESCRIBED_VEGETATION, PRESCRIBED_LAI
        self.SOIL_USES = ["Rainfed", "Forest", "Irrigated"]
        self.PRESCRIBED_VEGETATION = [fract + "_prescribed" for fract in self.SOIL_USES]
        self.PRESCRIBED_LAI = OrderedDict(zip(self.PRESCRIBED_VEGETATION[:], ['LAIOtherMaps', 'LAIForestMaps', 'LAIIrrigationMaps']))
        self.VEGETATION_LANDUSE = OrderedDict(zip(self.PRESCRIBED_VEGETATION[:], self.SOIL_USES[:]))
        
        
        
        self.LANDUSE_VEGETATION = OrderedDict([(v, [k]) for k, v in self.VEGETATION_LANDUSE.items()])
        self.LANDUSE_INPUTMAP = OrderedDict(zip(self.LANDUSE_VEGETATION.keys(), ["OtherFraction", "ForestFraction", "IrrigationFraction"]))

        if option['readNetcdfStack']:
            # get the extent of the maps from the netCDF template
            # and the modelling extent from the MaskMap
            # cutmap[] defines the MaskMap inside the precipitation map
            _ = CutMap(*mapattrNetCDF(binding['netCDFtemplate']))  # register cutmaps
        if option['writeNetcdfStack'] or option['writeNetcdf']:
            # if NetCDF is writen, the pr.nc is read to get the metadata
            # like projection
            _ = NetCDFMetadata(uuid.uuid4())  # init meta netcdf

        # ----------------------------------------

        # include all the hydrological modules
        self.misc_module = miscInitial(self)
        self.readmeteo_module = readmeteo(self)
        self.landusechange_module = landusechange(self)
        self.leafarea_module = leafarea(self)
        self.snow_module = snow(self)
        self.frost_module = frost(self)
        self.inflow_module = inflow(self)
        self.soil_module = soil(self)
        self.routing_module = routing(self)
        self.groundwater_module = groundwater(self)
        self.surface_routing_module = surface_routing(self)
        self.reservoir_module = reservoir(self)
        self.lakes_module = lakes(self)
        self.polder_module = polder(self)
        self.waterabstraction_module = waterabstraction(self)
        self.indicatorcalc_module = indicatorcalc(self)

        self.riceirrigation_module = riceirrigation(self)
        self.evapowater_module = evapowater(self)
        self.transmission_module = transmission(self)

        self.soilloop_module = soilloop(self)
        self.opensealed_module = opensealed(self)
        self.waterbalance_module = waterbalance(self)
        self.waterlevel_module = waterlevel(self)
        self.structures_module = structures(self)

        self.prescribed_vegetation = self.epic_settings.prescribed_vegetation
        self.interactive_vegetation = []
        if option.get('cropsEPIC'):
            self.prescribed_vegetation = self.epic_settings.prescribed_vegetation
            if int(binding["DtSec"]) != 86400:
                raise Exception("EPIC runs only using daily time steps!")
            from EPIC_modules.EPIC_main import EPIC_main # EPIC: agriculture simulator
            self.crop_module = EPIC_main(self) # EPIC: agriculture simulation
            if option["allIrrigIsEPIC"]: # the whole irrigated cropland is simulated by EPIC: remove 'Irrigated_prescribed' soil fraction
                self.prescribed_vegetation.remove('Irrigated_prescribed') # (also removed from PRESCRIBED_VEGETATION)
                self.epic_settings.vegetation_landuse.pop('Irrigated_prescribed')
                self.epic_settings.landuse_vegetation['Irrigated'] = []
                self.epic_settings.prescribe_lai.pop('Irrigated_prescribed')
            self.epic_settings.vegetation_landuse.update(self.crop_module.crop2landuse.to_dict()) # add EPIC crops to dictionary mapping vegetation fractions to land use types
            self.epic_settings.landuse_vegetation["Rainfed"] += self.crop_module.rainfed_crops.tolist()
            self.epic_settings.landuse_vegetation["Irrigated"] += self.crop_module.irrigated_crops.tolist()
            self.interactive_vegetation += self.crop_module.simulated_crops.tolist()
        self.vegetation = self.prescribed_vegetation + self.interactive_vegetation

        # --------------------------------------

        # include stateVar modules
        self.stateVar_module = stateVar(self)

        # run intial misc to get all global variables
        self.misc_module.initial()

        # include output of tss and maps
        self.output_module = outputTssMap(self)

        self.ReportSteps = report_steps['rep']

        self.landusechange_module.initial()

        self.snow_module.initial()
        self.frost_module.initial()
        self.leafarea_module.initial()
        self.soilloop_module.initial()

        self.soil_module.initial()
        self.routing_module.initial()

        self.groundwater_module.initial()
        self.waterlevel_module.initial()

        self.inflow_module.initial()
        self.surface_routing_module.initial()

        self.reservoir_module.initial()
        self.lakes_module.initial()
        self.polder_module.initial()

        self.transmission_module.initial()
        self.output_module.initial()

        self.structures_module.initial()
        # Structures such as reservoirs and lakes are modelled by interrupting the channel flow paths

        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------

        self.routing_module.initialSecond()
        # CHANNEL INITIAL SPLIT UP IN SECOND CHANNEL
        self.surface_routing_module.initialSecond()

        self.evapowater_module.initial()
        self.riceirrigation_module.initial()
        self.waterabstraction_module.initial()
        self.indicatorcalc_module.initial()

        self.waterbalance_module.initial()
        # calculate initial amount of water in the catchment

        if option.get('cropsEPIC'):
            self.crop_module.initial() # EPIC: agriculture simulator

        # debug start
        if flags['debug']:
            # Print value of variables after initialization (from state files)
            nomefile = 'Debug_init_' + str(self.currentStep + 1) + '.txt'
            ftemp1 = open(nomefile, 'w+')
            nelements = len(self.ChanM3)
            for i in range(0, nelements - 1):
                if hasattr(self, 'CrossSection2Area'):
                    print(i, self.TotalCrossSectionArea[i], self.CrossSection2Area[i], self.ChanM3[i], self.Chan2M3Kin[i], file=ftemp1)
                else:
                    print(i, self.TotalCrossSectionArea[i], self.ChanM3[i], file=ftemp1)

            ftemp1.close()

# ====== INITIAL ================================
    def initial(self):
        """ Initial part of LISFLOOD
            calls the initial part of the hydrological modules
        """
        # ----------------------------------------------------------------------
        # Perturbe the states
        #self.groundwater_module.var.UpperZoneK = perturbState(self.groundwater_module.var.UpperZoneK, method = "normal", minVal=0, maxVal=100, mu=self.groundwater_module.var.UpperZoneK, sigma=0.05, spatial=False)
        #self.groundwater_module.var.UZ = perturbState(self.groundwater_module.var.UZ, method = "normal", minVal=0, maxVal=100, mu=10, sigma=3, spatial=False, single=False)
        pass


# ====== METHODS ================================

    @property
    def num_pixel(self):
        """"""
        return self.maskinfo.info.mapC[0]
        # return maskinfo['mapC'][0]

    @property
    def dim_pixel(self):
        """"""
        return ("pixel", np.arange(self.num_pixel))

    @property
    def dim_vegetation(self):
        """"""
        return ("vegetation", self.vegetation[:])

    @property
    def dim_landuse(self):
        """"""
        return ("landuse", self.epic_settings.soil_uses[:])

    @property
    def dim_runoff(self):
        """"""
        return ("runoff", ["Other", "Forest", "Direct"])

    @property
    def coord_landuse(self):
        """"""
        return OrderedDict([self.dim_landuse, self.dim_pixel])

    @property
    def coord_vegetation(self):
        """"""
        return OrderedDict([self.dim_vegetation, self.dim_pixel])

    @property
    def coord_prescribed_vegetation(self):
        """"""
        return OrderedDict([("vegetation", self.prescribed_vegetation[:]), self.dim_pixel])

    def get_landuse_and_indexes_from_vegetation_epic(self, veg):
        """"""
        #iveg = self.coord_vegetation['vegetation'].index(veg)
        iveg = self.vegetation.index(veg)
        landuse = self.epic_settings.vegetation_landuse[veg] # landuse corresponding to vegetation fraction
        #ilanduse = self.coord_landuse['landuse'].index(landuse)
        ilanduse = self.epic_settings.soil_uses.index(landuse)
        return iveg,ilanduse,landuse

    def get_landuse_and_indexes_from_vegetation_GLOBAL(self, veg):
        """"""
        iveg = self.vegetation.index(veg)
        landuse = self.VEGETATION_LANDUSE[veg] # landuse corresponding to vegetation fraction
        ilanduse = self.SOIL_USES.index(landuse)
        return iveg,ilanduse,landuse

    def get_indexes_from_landuse_and_veg_list_GLOBAL(self, landuse, veg_list):
        """"""
        ilanduse = self.SOIL_USES.index(landuse)
        iveg_list_pres = []
        iveg_list = []
        for veg in veg_list:
            iveg_list_pres.append(self.PRESCRIBED_VEGETATION.index(veg))
            iveg_list.append(self.vegetation.index(veg))
        return iveg_list,iveg_list_pres,ilanduse

    def allocateVariableAllVegetation(self, dtype=float):
        """Allocate xarray.DataArray filled by 0 with dimensions 'vegetation' and 'pixel'. It covers all vegetation types (including EPIC crops, if simulated)."""
        return self.allocateDataArray(self.coord_vegetation, dtype)

    def allocateDataArray(self, dimensions, dtype=float):
        """Allocate xarray.DataArray filled by 0 with input dimensions.
           Argument 'dimensions' is a list of tuples of the type ('dimension name', coordinate list/array)."""
        coords = OrderedDict(dimensions)
        option = self.settings.options
        if option.get('cropsEPIC'):
            return xr.DataArray(np.zeros([len(v) for v in coords.values()], dtype), coords=coords, dims=coords.keys())
        else:
            return NumpyModified(np.zeros([len(v) for v in coords.values()], dtype), dims=coords.keys())

    def initialiseVariableAllVegetation(self, name, coords=None):
        """Load a DataArray from a model output netCDF file (typycally an end map).
        This function allows reading netCDF variables with more than 3 dimensions (time, y, x) into a xarray.DataArray.
        The coords argument is used if name does not point to a netCDF file: coordinates to allocate the DataArray before assigning a default value."""
        file_path = self.settings.binding[name]
        if os.path.exists(file_path):
            variable = ".".join(os.path.split(file_path)[1].split(".")[:-1]) # the outer join allows treating variables name of the type 'variable.end'
            try:
                with xr.open_dataset(file_path)[variable] as nc:
                    coords = [(dim, nc.coords[dim].values) for dim in nc.dims[:-2]] + [self.dim_pixel]
                    values = nc.values[...,~self.maskinfo.info.mask]  # maskinfo["mask"]
            except:
                raise Exception("{} must be a netCDF file! Check the input {} in the settings file!".format(file_path, name))
        else:
            try:
                if coords is None:
                    coords = self.coord_vegetation
                values = float(file_path)
            except:
                raise Exception("The input {} must be either the path to a netCDF file or a string that can be converted to a float!".format(name))
        output = self.allocateDataArray(coords)
        output[:] = values
        return output

    def defsoil(self, name_1, name_2=None, name_3=None, coords=None):
        """Load input/parameter static maps for 3 pixel fractions and arrange them in a xarray.DataArray.
           Default fractions: Rainfed, Forest, Irrigated. Other fractions can be specified via the coords keyword argument """
        if coords is None:
            coords = self.coord_landuse
        data = self.allocateDataArray(coords)
        values_1 = readInputWithBackup(name_1)
        option = self.settings.options
        if option.get('cropsEPIC'):
            labels = list(coords.values())[0]
            data.loc[labels[0],:] = values_1
            data.loc[labels[1],:] = readInputWithBackup(name_2, values_1)
            data.loc[labels[2],:] = readInputWithBackup(name_3, values_1)
        else:
            if (list(coords.keys())[0]=="landuse") or (list(coords.keys())[0]=="runoff"):
                data.values[0][:] = values_1
                data.values[1][:] = readInputWithBackup(name_2, values_1)
                data.values[2][:] = readInputWithBackup(name_3, values_1)
            else:
                raise Exception("Coords key not found!")
        return data

    def deffraction(self, variable):
        """Weighted sum over the soil fractions of each pixel"""
        ax_veg = variable.dims.index("vegetation")
        return _vegSum(ax_veg, variable.values, self.SoilFraction.values)



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""
from __future__ import print_function, absolute_import

import cftime
from future.utils import listitems

from nine import range

import uuid
import warnings
import re
import time as xtime
import sys
import datetime
import os
import pickle
from bisect import bisect_left

import pcraster
from pcraster import Scalar, numpy2pcr, Nominal, setclone, Boolean, pcr2numpy
from netCDF4 import num2date, date2num
import numpy as np
#import xarray as xr

from .zusatz import iterOpenNetcdf, iterReadPCRasterMap, iterSetClonePCR, checkmap
from .settings import (calendar_inconsistency_warning, get_calendar_type, calendar, MaskAttrs, CutMap, NetCDFMetadata,
                       LisSettings, MaskInfo)
from .errors import LisfloodWarning, LisfloodError
from .decorators import Cache

# modified numpy class to ensure code compatibility with EPIC (that uses XArray in computations)
# while keeping higher performance when EPIC modules are not needed (use of only numpy arrays in computations)
class NumpyModified(np.ndarray):
    obj_dims = []
    def __new__(cls, input_array, dims):
        obj = np.asarray(input_array).view(cls)
        obj.obj_dims = list(dims)
        return obj

    @property
    def values(self):
        return self
    
    @property
    def dims(self):
        return self.obj_dims


def defsoil(name1, name2=None, name3=None):
    """ loads 3 array in a list
    """
    if isinstance(name1, str):
        in1 = loadmap(name1)
    else:
        in1 = name1

    if name2 is None:
        in2 = in1
    else:
        if isinstance(name2, str):
            in2 = loadmap(name2)
        else:
            in2 = name2

    if name3 is None:
        in3 = in1
    else:
        if isinstance(name3, str):
            in3 = loadmap(name3)
        else:
            in3 = name3

    return [in1, in2, in3]


def readInputWithBackup(name, values_if_failure=None):
    """Read input map/value if name is not None; else the provided backup values are used"""
    if name is None:
        return values_if_failure
    else:
        try:
            return loadmap(name)
        except:
            return name


def valuecell(mask, coordx, coordstr):
    """
    to put a value into a pcraster map -> invert of cellvalue
    pcraster map is converted into a numpy array first
    """
    coord = []
    for xy in coordx:
        try:
            coord.append(float(xy))
        except:
            msg = "Gauges: " + xy + " in " + coordstr + " is not a coordinate"
            raise LisfloodError(msg)

    null = np.zeros((pcraster.clone().nrRows(), pcraster.clone().nrCols()))
    null[null == 0] = -9999

    for i in range(int(len(coord) / 2)):
        col = int(
            (coord[i * 2] - pcraster.clone().west()) / pcraster.clone().cellSize())
        row = int(
            (pcraster.clone().north() - coord[i * 2 + 1]) / pcraster.clone().cellSize())
        #if col >= 0 and row >= 0 and col < pcraster.clone().nrCols() and row < pcraster.clone().nrRows():
        if col >= 0 and row >= 0 and col < pcraster.clone().nrCols() and row < pcraster.clone().nrRows():
            null[row, col] = i + 1
        else:
            msg = "Coordinates: " + str(coord[i * 2]) + ',' + str(
                coord[i * 2 + 1]) + " to put value in is outside mask map - col,row: " + str(col) + ',' + str(row)
            raise LisfloodError(msg)

    map = numpy2pcr(Nominal, null, -9999)
    return map


def mapattrNetCDF(name):
    """
    get the map attributes like col, row etc from a ntcdf map
    and define the rectangular of the mask map inside the netcdf map
    """
    filename = os.path.splitext(name)[0] + '.nc'
    nf1 = iterOpenNetcdf(filename, "Checking netcdf map \n", 'r')
    spatial_dims = ('x', 'y') if 'x' in nf1.variables else ('lon', 'lat')
    x1, x2, y1, y2 = [nf1.variables[v][j] for v in spatial_dims for j in (0, 1)]
    nf1.close()
    maskattrs = MaskAttrs.instance()

    cell_x = np.abs(x2 - x1)
    cell_y = np.abs(y2 - y1)
    check_x = maskattrs['cell'] - cell_x # this must be same precision as pcraster.clone().cellsize()
    check_y = maskattrs['cell'] - cell_y # this must be same precision as pcraster.clone().cellsize()

    if abs(check_x) > 10**-5 or abs(check_y) > 10**-5:
        raise LisfloodError("Cell size different in maskmap {} and {}".format(
            LisSettings.instance().binding['MaskMap'], filename)
        )
    half_cell = maskattrs['cell'] / 2.
    x = x1 - half_cell  # |
    y = y1 + half_cell  # | coordinates of the upper left corner of the input file upper left pixel
    cut0 = int(np.abs(maskattrs['x'] - x) / cell_x)
    cut1 = cut0 + maskattrs['col']
    cut2 = int(np.abs(maskattrs['y'] - y) / cell_y)
    cut3 = cut2 + maskattrs['row']
    return cut0, cut1, cut2, cut3  # input data will be sliced using [cut0:cut1,cut2:cut3]


def loadsetclone(name):
    """ Load 'MaskMap' and set as clone
        
    :param name: name of the key in Settings.xml containing path and name of mask map as string
    :return: map: mask map (False=include in modelling; True=exclude from modelling) as pcraster
    """
    settings = LisSettings.instance()
    binding = settings.binding
    flags = settings.flags
    filename = os.path.normpath(binding[name])
    if not os.path.exists(filename):
        raise LisfloodError('File not existing: {}'.format(filename))
    coord = filename.split()    # returns a list of all the words in the string
    if len(coord) == 5:
        # changed order of x, y i- in setclone y is first in Lisflood
        # settings x is first
        # setclone row col cellsize xupleft yupleft
        try:
            setclone(int(coord[1]), int(coord[0]), float(coord[2]), float(coord[3]), float(coord[4]))   # CM: pcraster
        except:
            rem = "["+str(coord[0])+" "+ str(coord[1])+" "+ str(coord[2])+" "+ str(coord[3])+" "+str(coord[4])+"]"
            msg = "Maskmap: " + rem + \
                  " are not valid coordinates (col row cellsize xupleft yupleft)"
            raise LisfloodError(msg)
        mapnp = np.ones((int(coord[1]), int(coord[0])))
        map_out = numpy2pcr(Boolean, mapnp, -9999)
    elif len(coord) == 1:
        # read information on clone map from map (pcraster or netcdf)
        try:
            # try to read a pcraster map
            iterSetClonePCR(filename)
            map_out = pcraster.boolean(iterReadPCRasterMap(filename))
            flagmap = True
            mapnp = pcr2numpy(map_out, np.nan)
        except Exception as e:
            # FIXME manage exceptions and print type of error
            # print(str(e))
            # print(type(e))
            # try to read a netcdf file
            filename = os.path.splitext(binding[name])[0] + '.nc'
            nf1 = iterOpenNetcdf(filename, "", "r")
            value = listitems(nf1.variables)[-1][0]  # get the last variable name
            nr_rows, nr_cols = nf1.variables[value].shape  # just use shape to know rows and cols...
            if 'x' in nf1.variables:
                x1 = nf1.variables['x'][0]
                x2 = nf1.variables['x'][-1]
                y1 = nf1.variables['y'][0]
            else:
                x1 = nf1.variables['lon'][0]
                x2 = nf1.variables['lon'][-1]
                y1 = nf1.variables['lat'][0]

            cell_size = np.abs(x2 - x1)/(nr_cols - 1)
            x = x1 - cell_size / 2
            y = y1 + cell_size / 2
            mapnp = np.array(nf1.variables[value][0:nr_rows, 0:nr_cols])
            nf1.close()
            # setclone  row col cellsize xupleft yupleft
            setclone(nr_rows, nr_cols, cell_size, x, y)
            map_out = numpy2pcr(Boolean, mapnp, 0)
            flagmap = True

        if flags['checkfiles']:
            checkmap(name, filename, map_out, flagmap, 0)
    else:
        raise LisfloodError("Maskmap: {} is not a valid mask map nor valid coordinates".format(name))
    _ = MaskAttrs(uuid.uuid4())  # init maskattrs
    # put in the ldd map
    # if there is no ldd at a cell, this cell should be excluded from modelling
    ldd = loadmap('Ldd', pcr=True)
    # convert ldd to numpy
    maskldd = pcr2numpy(ldd, np.nan)
    # convert numpy map to 8bit
    maskarea = np.bool8(mapnp)
    # compute mask (pixels in maskldd AND maskarea)
    mask = np.logical_not(np.logical_and(maskldd, maskarea))
    _ = MaskInfo(mask, map_out)  # MaskInfo init here

    if flags['nancheck']:
        nanCheckMap(ldd, binding['Ldd'], 'Ldd')
    return map_out


def compressArray(map, pcr=True, name=None):
    maskinfo = MaskInfo.instance()
    if pcr:
        mapnp = pcr2numpy(map,np.nan)
        mapnp1 = np.ma.masked_array(mapnp, maskinfo.info.mask)
    else:
        mapnp1 = np.ma.masked_array(map, maskinfo.info.mask)
    mapC = np.ma.compressed(mapnp1)

    if name is not None:
        if np.max(np.isnan(mapC)):
            msg = name + " has less valid pixels than area or ldd \n"
            raise LisfloodError(msg)
            # test if map has less valid pixel than area.map (or ldd)
    return mapC.astype(float)


def decompress(map):
    maskinfo = MaskInfo.instance()
    dmap = maskinfo.info.maskall.copy()
    dmap[~maskinfo.info.maskflat] = map[:]
    dmap = dmap.reshape(maskinfo.info.shape)
    # check if integer map (like outlets, lakes etc)
    try:
        checkint = str(map.dtype)
    except:
        checkint = None

    if checkint in ("int16", "int32", "int64"):
        dmap[dmap.mask] = -9999
        map = numpy2pcr(Nominal, dmap, -9999)
    elif checkint == "int8":
        dmap[dmap < 0] = -9999
        map = numpy2pcr(Nominal, dmap, -9999)
    else:
        dmap[dmap.mask] = -9999
        map = numpy2pcr(Scalar, dmap, -9999)
    return map


def makenumpy(map):
    if not isinstance(map, np.ndarray):
        maskinfo = MaskInfo.instance()
        out = np.empty(maskinfo.info.mapC)
        out.fill(map)
        return out
    else:
        return map


def loadmap(*args, **kwargs):
    settings = LisSettings.instance()
    binding = settings.binding

    if binding['MapsCaching'] == "True":
        # get path to map file to make sure it's unique in the cache
        if len(args) > 0:
            name = args[0]
        else:
            name = kwargs['name']
        value = binding[name]
        kwargs['value'] = value
        data = loadmap_cached(*args, **kwargs)
    else:
        data = loadmap_base(*args, **kwargs)
    
    return data

@Cache
def loadmap_cached(*args, **kwargs):
    return loadmap_base(*args, **kwargs)


def loadmap_base(name, pcr=False, lddflag=False, timestampflag='exact', averageyearflag=False, value=None):
    """ Load a static map either value or pcraster map or netcdf (single or stack)
    
    Load a static map either value or pcraster map or netcdf (single or stack)
    If a netCDF stack is loaded, map is read according to timestepInit date (i.e. model time step). If timestepInit is a
    step number, step number is converted to date (referred to CalendarDayStart in settings.xml). Then date is used to
    read time step from netCDF file.
    if timestampflag = 'closest' and loadmap is reading a NetCDF stack, the timestep with the closest timestamp will be
    loaded if the exact one is not available.
    
    :param name: name of key in Settings.xml input file containing path and name of the map file (as string)
    :param pcr: flag for output maps in pcraster format 
    :param lddflag: flag for local drain direction map (CM??)
    :param timestampflag: look for exact time stamp in netcdf file ('exact') or for the closest (left) time stamp available ('closest')
    :param averageyearflag: if True, use "average year" netcdf file over the entire model simulation period
    :return: map or mapC
    :except: pcr: maps must have the same size of clone.map
             netCDF: time step timestepInit must be included into the stack 
    """
    # name of the key in Settimgs.xml file containing path and name of the map file
    settings = LisSettings.instance()
    binding = settings.binding
    flags = settings.flags
    if value is None:
        value = binding[name]
    # path and name of the map file
    filename = value
    load = False
    pcrmap = False
    # try reading in PCRaster map format
    try:
        # try reading constant value
        mapC = float(value)
        flagmap = False
        load = True
        if pcr: map=mapC
    except ValueError:
        try:
            # try reading pcraster map exploiting the iterAccess class
            map = iterReadPCRasterMap(value)
            flagmap = True
            load = True
            pcrmap = True
        except:
            load = False

    if load and pcrmap:
        #map is loaded and it is in pcraster format
        try:
            # test if map is same size as clone map, if not it will make an error
           test = pcraster.scalar(map) + pcraster.scalar(map)
        except:
           raise LisfloodError("{} might be of a different size than clone size".format(value))
    # if failed before try reading from netCDF map format
    if not load:
        # read a netcdf  (single one not a stack)
        filename = os.path.splitext(value)[0] + '.nc'
        # get mapextend of netcdf map and calculate the cutting
        cut0, cut1, cut2, cut3 = mapattrNetCDF(filename)
        # load netcdf map but only the rectangle needed
        nf1 = iterOpenNetcdf(filename, "", 'r')
        # Only one variable must be present in netcdf files
        num_dims = 3 if 'time' in nf1.variables else 2
        varname = [v for v in nf1.variables if len(nf1.variables[v].dimensions) == num_dims][0]
        if not settings.timestep_init:
            # if timestep_init is missing, read netcdf as single static map
            mapnp = nf1.variables[varname][cut2:cut3, cut0:cut1]
        else:
            if 'time' in nf1.variables:
                # read a netcdf  (stack) - state files
                # get information from netCDF stack
                t_steps = nf1.variables['time'][:]  # get values for timesteps ([  0.,  24.,  48.,  72.,  96.])
                t_unit = nf1.variables['time'].units  # get unit (u'hours since 2015-01-01 06:00:00')
                t_cal = get_calendar_type(nf1)
                # get year from time unit in case average year is used
                if averageyearflag:
                    # get date of the first step in netCDF file containing average year values
                    first_date = num2date(t_steps[0],t_unit,t_cal)
                    # get year of the first step in netCDF file containing average year values
                    t_ref_year = first_date.year

                # select timestep to use for reading from netCDF stack based on timestep_init (state file time step)
                timestepI = calendar(settings.timestep_init, binding['calendar_type'])
                if isinstance(timestepI, (datetime.datetime, cftime.DatetimeProlepticGregorian, cftime.real_datetime)):
                    #reading dates in XML settings file
                    # get step id number in netCDF stack for timestepInit date
                    if averageyearflag:
                        #if using an average year don't care about the year in timestepIDate and change it to the netCDF first time step year
                        try:
                            timestepI = timestepI.replace(year=t_ref_year)
                        except:
                            timestepI = timestepI.replace(day=28)
                            timestepI = timestepI.replace(year=t_ref_year)
                    timestepI = date2num(timestepI, nf1.variables['time'].units)
                else:
                    # reading step numbers in XML file
                    # timestepI = int(timestepI) -1
                    begin = calendar(binding['CalendarDayStart'])
                    DtSec = float(binding['DtSec'])
                    DtDay = DtSec / 86400.
                    # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)
                    # get date for step number timestepI (referred to CalendarDayStart)
                    timestepIDate = begin + datetime.timedelta(days=(timestepI - 1) * DtDay)
                    # get step id number in netCDF stack for step timestepInit
                    # timestepInit refers to CalenradDayStart
                    # timestepI now refers to first date in netCDF stack
                    if averageyearflag:
                        #using an average year, don't care about the year in timestepIDate and change it to the netCDF time unit year
                        try:
                            timestepIDate = timestepIDate.replace(year=t_ref_year)
                        except:
                            #if simulation year is leap and average year is not, switch 29/2 with 28/2
                            timestepIDate = timestepIDate.replace(day=28)
                            timestepIDate = timestepIDate.replace(year=t_ref_year)
                    timestepI = date2num(timestepIDate, units=t_unit, calendar=t_cal)

                if not(timestepI in nf1.variables['time'][:]):
                    if timestampflag == 'exact':
                        #look for exact time stamp when loading data
                        msg = "time step " + str(int(timestepI)+1)+" is not stored in "+ filename
                        raise LisfloodError(msg)
                    elif timestampflag == 'closest':
                        #get the closest value
                        timestepInew = takeClosest(t_steps, timestepI)
                        #set timestepI to the closest available time step in netCDF file
                        timestepI = timestepInew

                itime = np.where(nf1.variables['time'][:] == timestepI)[0][0]
                mapnp = nf1.variables[varname][itime, cut2:cut3, cut0:cut1]
            else:
                # read a netcdf (single one)
                mapnp = nf1.variables[varname][cut2:cut3, cut0:cut1]

        # masking
        try:
            maskinfo = MaskInfo.instance()
            mapnp.mask = maskinfo.info.mask
        except (KeyError, AttributeError):
            pass
        nf1.close()

        # if a map should be pcraster
        if pcr:
            # check if integer map (like outlets, lakes etc
            checkint = str(mapnp.dtype)
            if checkint == "int16" or checkint == "int32":
                mapnp[mapnp.mask] = -9999
                map = numpy2pcr(Nominal, mapnp, -9999)
            elif checkint == "int8":
                mapnp[mapnp < 0] = -9999
                map = numpy2pcr(Nominal, mapnp, -9999)
            else:
                mapnp[np.isnan(mapnp)] = -9999
                map = numpy2pcr(Scalar, mapnp, -9999)
            # if the map is a ldd
            if lddflag:
                map = pcraster.ldd(pcraster.nominal(map))
        else:
            mapC = compressArray(mapnp, pcr=False, name=filename)
        flagmap = True

    # pcraster map but it has to be an array
    if pcrmap and not pcr:
        mapC = compressArray(map, name=filename)

    if flags['checkfiles']:
        print(name, filename)
        if flagmap == False:
            checkmap(name, filename, mapC, flagmap, 0)
        elif pcr:
            checkmap(name, filename, map, flagmap, 0)
        else:
            print(name, mapC.size)
            if mapC.size > 0:
                map= decompress(mapC)
                checkmap(name, filename, map, flagmap, 0)
    if pcr:
        if flags['nancheck'] and name != 'Ldd':
            nanCheckMap(map, filename, name)
        return map
    elif isinstance(mapC, np.ndarray):
        return mapC.astype(float)
    else:
        if flags['nancheck'] and name != 'Ldd':
            nanCheckMap(mapC, filename, name)
        return mapC


def takeClosest(myList, myNumber):
    """ Returns the closest left value to myNumber in myList
    
    Assumes myList is sorted. Returns closest left value to myNumber.
    If myList is sorted in raising order, it returns the closest smallest value.
    https://stackoverflow.com/questions/12141150/from-list-of-integers-get-number-closest-to-a-given-value
    
    :param myList: list of ordered values
    :param myNumber: number to be searche in myList
    :return: closest left number to myNumber in myList
    """
    pos = bisect_left(myList, myNumber)
    if pos == 0:
        return myList[0]
    if pos == len(myList):
        return myList[-1]
    before = myList[pos - 1]
    # after = myList[pos]
    # if after - myNumber < myNumber - before:
    #    return after
    # else:
    return before


def loadLAI(value, pcrvalue, i, pcr=False):
    """
    load Leaf are map stacks  or water use maps stacks
    """
    pcrmap = False
    settings = LisSettings.instance()
    flags = settings.flags

    try:
        map = iterReadPCRasterMap(pcrvalue)
        filename = pcrvalue
        pcrmap = True
    except:
        filename = os.path.splitext(value)[0] + '.nc'
        # get mapextend of netcdf map
        # and calculate the cutting
        cut0, cut1, cut2, cut3 = mapattrNetCDF(filename)

        nf1 = iterOpenNetcdf(filename, "", 'r')
        # Only one variable must be present in netcdf files
        num_dims = 3 if 'time' in nf1.variables else 2
        varname = [v for v in nf1.variables if len(nf1.variables[v].dimensions) == num_dims][0]
        mapnp = nf1.variables[varname][i, cut2:cut3, cut0:cut1]
        nf1.close()
        mapC = compressArray(mapnp, pcr=False, name=filename)
        # mapnp[np.isnan(mapnp)] = -9999
        # map = numpy2pcr(Scalar, mapnp, -9999)
        # if check use a pcraster map
        if flags['checkfiles'] or pcr:
            map = decompress(mapC)
    if pcrmap: mapC = compressArray(map,name=filename)
    if flags['checkfiles']:
        checkmap(os.path.basename(pcrvalue), filename, map, True, 0)
    map_out = map if pcr else mapC
    if flags['nancheck']:
        nanCheckMap(map_out, filename, "'LAI*Maps' or 'WFractionMaps'")
    return map_out
    # if pcr:
    #     if flags['nancheck']:
    #         nanCheckMap(map, filename, "'LAI*Maps' or 'WFractionMaps'")
    #     return map
    # else:
    #     if flags['nancheck']:
    #         nanCheckMap(mapC, filename, "'LAI*Maps' or 'WFractionMaps'")
    #     return mapC


def readmapsparse(name, time, oldmap):
    """
    load stack of maps 1 at each timestamp in Pcraster format
    """
    filename = generateName(name, time)
    flags = LisSettings.instance().flags
    try:
        map = iterReadPCRasterMap(filename)
        find = 1
    except:
        find = 2
        if oldmap is None:
            for i in range(time - 1, 0, -1):
                altfilename = generateName(name, i)
                if os.path.exists(altfilename):
                    map = iterReadPCRasterMap(altfilename)
                    find = 1
                    # break
            if find == 2:
                msg = "no map in stack has a smaller time stamp than: " + filename
                raise LisfloodError(msg)
        else:
            map = oldmap
            if flags['loud']:
                s = " last_%s" % (os.path.basename(name))
                print(s)
    if flags['checkfiles']:
        checkmap(os.path.basename(name), filename, map, True, find)
    if flags['nancheck']:
        nanCheckMap(map, filename, name)
    mapC = compressArray(map,name=filename)
    return mapC


def readnetcdf(name, time, timestampflag='exact', averageyearflag=False):
    """ Read maps from netCDF stacks (forcings, fractions, water demand)

    Read maps from netCDF stacks (forcings, fractions, water demand).
    Maps are read by date, so stacks can start at every date also different from CalendarDayStart.
    Units for stacks can be different from model timestep.
    It can read sub-daily steps.
    timestampflag indicates whether to load data with the exact time stamp ('exact'), or the data with the closest time
    stamp when the exact one is not available ('closest').
    averageyearflag indicates whether to load data from a netcdf file containing one single "average" year (it's used for
    water demand and landuse changes in time).

    :param name: string containing path and name of netCDF file to be read
    :param time: current simulation timestep of the model as integer number (referred to CalendarStartDay)
    :param timestampflag: look for exact time stamp in netcdf file ('exact') or for the closest (left) time stamp available ('closest')
    :param averageyearflag: if True, use "average year" netcdf file over the entire model simulation period
    :returns: content of netCDF map for timestep "time" (mapC)
    :except: if current simulation timestep is not stored in the stack, it stops with error message (if timestampflag='exact')
    """

    filename = name + ".nc" if not name.endswith('.nc') else name
    nf1 = iterOpenNetcdf(filename, "Netcdf map stacks: \n", "r")

    # read information from netCDF file
    variable_name = [k for k in nf1.variables if len(nf1.variables[k].dimensions) == 3][0]  # get the variable with 3 dimensions (variable order not relevant)
    t_steps = nf1.variables['time'][:]    # get values for timesteps ([  0.,  24.,  48.,  72.,  96.])
    t_unit = nf1.variables['time'].units  # get unit (u'hours since 2015-01-01 06:00:00')
    t_cal = get_calendar_type(nf1)
    # CM: get year from time unit in case average year is used
    if averageyearflag:
        # CM: get date of the first step in netCDF file containing average year values
        first_date = num2date(t_steps[0], t_unit, t_cal)
        # CM: get year of the first step in netCDF file containing average year values
        t_ref_year = first_date.year
    settings = LisSettings.instance()
    binding = settings.binding
    flags = settings.flags
    begin = calendar(binding['CalendarDayStart'], binding['calendar_type'])
    dt_sec = float(binding['DtSec'])
    dt_day = float(dt_sec / 86400)
    # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)

    # get date of current simulation step
    currentDate = calendar(time, binding['calendar_type'])
    if type(currentDate) is not datetime.datetime:
        currentDate = begin + datetime.timedelta(days=(currentDate - 1) * dt_day)

    # if reading from an average year NetCDF stack, ignore the year in current simulation date and change it to the netCDF time unit year
    if averageyearflag:
        try:
            currentDate = currentDate.replace(year=t_ref_year)
        except:
            # CM: if simulation year is leap and average year is not, switch 29/2 with 28/2
            currentDate = currentDate.replace(day=28)
            currentDate = currentDate.replace(year=t_ref_year)

    # get timestep in netCDF file corresponding to current simulation date
    current_ncdf_step = date2num(currentDate, units=t_unit, calendar=t_cal)

    # read netCDF map
    if not (current_ncdf_step in t_steps):
        if (timestampflag == 'exact'):
            # look for exact time stamp when loading data
            msg = "Date " + str(currentDate) + " not stored in " + filename
            raise LisfloodError(msg)
        elif (timestampflag == 'closest'):
            # CM: get the closest value
            current_ncdf_step_new = takeClosest(t_steps, current_ncdf_step)
            # CM: set current_ncdf_step to the closest available time step in netCDF file
            current_ncdf_step = current_ncdf_step_new

    # get index of timestep in netCDF file corresponding to current simulation date
    current_ncdf_index = np.where(t_steps == current_ncdf_step)[0][0]

    # crop map if subcatchment
    cutmap = CutMap.instance()
    mapnp = nf1.variables[variable_name][current_ncdf_index, cutmap.cuts[2]:cutmap.cuts[3], cutmap.cuts[0]:cutmap.cuts[1]]
    nf1.close()

    mapC = compressArray(mapnp, pcr=False, name=filename)
    if flags['checkfiles']:
        timename = os.path.basename(name) + str(time)
        checkmap(timename, filename, decompress(mapC), True, 1)
    if flags['nancheck']:
        nanCheckMap(mapC, filename, name)
    return mapC


def readnetcdfsparse(name, time, oldmap):
    """
    NO LONGER USED
    load stack of maps 1 at each timestamp in Netcdf format
    """
    try:
        mapC = readnetcdf(name, time)
        find = 1
        # print name+str(time)+"   "+str(find)
    except:
        find = 2
        if oldmap is None:
            for i in range(time - 1, 0, -1):
                try:
                    mapC = readnetcdf(name, i)
                    find = 1
                    break
                except:
                    pass
                # print name+"   "+str(time)+"   "+str(find)+"   "+str(i)
            if find == 2:
                msg = "no map in stack has a smaller time stamp than: " + str(time)
                raise LisfloodError(msg)
        else:
            settings = LisSettings.instance()
            flags = settings.flags
            mapC = oldmap
            if flags['loud']:
                s = " last_" + (os.path.basename(name)) + str(time)
                # print s,
    return mapC


def checknetcdf(name, start, end):
    """ Check available time steps in netCDF input file
    
    Check available timesteps in netCDF file. Get first and last available timestep in netCDF file and compare with
    first and last computation timestep of the model.
    It can use sub-daily steps.
    
    :param name: string containing path and name of netCDF file
    :param start: initial date or step number of model simulation
    :param end: final date or step of model simulation  
    :return: none
    :raises Exception: stop if netCDF maps do not cover simulation time period
    """
    settings = LisSettings.instance()
    binding = settings.binding
    filename = name + ".nc"
    nf1 = iterOpenNetcdf(filename, "Netcdf map stacks: \n", "r")

    # read information from netCDF file
    t_steps = nf1.variables['time'][:]    # get values for timesteps ([  0.,  24.,  48.,  72.,  96.])
    t_unit = nf1.variables['time'].units  # get unit (u'hours since 2015-01-01 06:00:00')
    t_cal = get_calendar_type(nf1)
    if t_cal != binding['calendar_type']:
        warnings.warn(calendar_inconsistency_warning(filename, t_cal, binding['calendar_type']))

    # get date of first available timestep in netcdf file
    date_first_step_in_ncdf = num2date(t_steps[0], units=t_unit, calendar=t_cal)
    # get date of last available timestep in netcdf file
    date_last_step_in_ncdf = num2date(t_steps[-1], units=t_unit, calendar=t_cal)

    nf1.close()
    #calendar date start (CalendarDayStart)
    begin = calendar(binding['CalendarDayStart'], binding['calendar_type'])
    DtSec = float(binding['DtSec'])
    DtDay = DtSec / 86400.
    # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)

    date_first_sim_step = calendar(start, binding['calendar_type'])
    if not isinstance(date_first_sim_step, (datetime.datetime, cftime.real_datetime)):
        date_first_sim_step = begin + datetime.timedelta(days=(date_first_sim_step - 1) * DtDay)
    if (date_first_sim_step < date_first_step_in_ncdf):
        msg = "First simulation time step is before first time step in netCDF input data file \n" \
              "File name: "+ filename +"\n" \
              "netCDF start date: "+ date_first_step_in_ncdf.strftime('%d/%m/%Y %H:%M') +"\n" \
              "simulation start date: "+ date_first_sim_step.strftime('%d/%m/%Y %H:%M')
        raise LisfloodError(msg)

    date_last_sim_step = calendar(end, binding['calendar_type'])
    if not isinstance(date_last_sim_step, (datetime.datetime, cftime.real_datetime)):
    # if type(date_last_sim_step) is not datetime.datetime:
        date_last_sim_step = begin + datetime.timedelta(days=(date_last_sim_step - 1) * DtDay)
    if (date_last_sim_step > date_last_step_in_ncdf):
        msg = "Last simulation time step is after last time step in netCDF input data file \n" \
              "File name: " + filename +"\n" \
              "netCDF last date: " + date_last_step_in_ncdf.strftime('%d/%m/%Y %H:%M') +"\n" \
              "simulation last date: " + date_last_sim_step.strftime('%d/%m/%Y %H:%M')
        raise LisfloodError(msg)
    return


def generateName(name, time):
    """Returns a filename based on the name and time step passed in.
    The resulting name obeys the 8.3 DOS style format. The time step
    will be added to the end of the filename and be prepended by 0's if
    needed.
    The time step normally ranges from [1, nrTimeSteps].
    The length of the name should be max 8 characters to leave room for
    the time step.
    The name passed in may contain a directory name.
    See also: generateNameS(), generateNameST()
    """
    head, tail = os.path.split(name)
    if re.search("\.", tail):
        msg = "File extension given in '" + name + "' not allowed"
        raise LisfloodError(msg)
    if len(tail) == 0:
        msg = "No filename specified"
        raise LisfloodError(msg)
    if len(tail) > 8:
        msg = "Filename '" + name + "' must be shorter than 8 characters"
        raise LisfloodError(msg)
    if time < 0:
        msg = "Timestep must be larger than 0"
        raise LisfloodError(msg)

    nr = "%d" % (time)
    space = 11 - (len(tail) + len(nr))
    assert space >= 0
    result = "%s%s%s" % (tail, space * "0", nr)
    result = "%s.%s" % (result[:8], result[8:])
    assert len(result) == 12
    return os.path.join(head, result)


def dumpObject(name, var, num):
  path1 = os.path.join(str(num), 'stateVar',name)
  file_object1 = open(path1, 'w')
  pickle.dump(var, file_object1)
  file_object1.close()


def loadObject(name, num):
  path1 = os.path.join(str(num), 'stateVar', name)
  filehandler1 = open(path1, 'r')
  #read a string from the open file object file and interpret it as a pickle data stream, rec
  var = pickle.load(filehandler1)
  filehandler1.close()
  return(var)


def dumpPCRaster(name, var, num):
  path1 = os.path.join(str(num), 'stateVar',name)
  pcraster.report(var, path1)


def loadPCRaster(name, num):
  path1 = os.path.join(str(num), 'stateVar',name)
  var = iterReadPCRasterMap(path1)
  return(var)

def perturbState(var, method = "normal", minVal=0, maxVal=1, mu=0, sigma=1, spatial=True, single=True):
  try:
    numVals = len(var)
  except:
    numVals = 1
  if method == "normal":
    if spatial:
      domain = len(var[0])
      out = var
      for i in range(numVals):
        out[i] = np.minimum(np.maximum(np.random.normal(mu, sigma, domain), minVal), maxVal)
    else:
      if single:
        out = np.minimum(np.maximum(np.random.normal(mu, sigma, numVals), minVal), maxVal)
      else:
        out = list(np.minimum(np.maximum(np.random.normal(mu, sigma, numVals), minVal), maxVal))
  if method == "uniform":
    if spatial:
      domain = len(var[0])
      out = var
      for i in range(numVals):
        out[i] = np.random.uniform(minVal, maxVal, domain)
    else:
      if single:
        out = np.random.uniform(minVal, maxVal, numVals)
      else:
        out = list(np.random.uniform(minVal, maxVal, numVals))
  return(out)


def read_tss_header(tssfilename):
    """ Read header of a tss file (used in inflow)
        :param tssfilename  path and name of the tss
        :returns outlets_id  list of column names in tss file 
    """
    with open(tssfilename) as fp:
        rec = fp.readline()
        if rec.split()[0] == 'timeseries':
            # LISFLOOD tss file with header
            # get total number of outlets
            outlets_tot_number = int(fp.readline())
            fp.readline()
            outlets_id = []
            for i in range(0, outlets_tot_number - 1):
                rec = fp.readline()
                rec = int(rec.strip())
                outlets_id.append(rec)  #Lisflood ID code for output points
            # read tss data
            # tssdata = pd.read_table(tssfilename, delim_whitespace=True, header=None, names=outlets_id, index_col=0,
            #                        skiprows=outlets_tot_number + 2)

        else:
            # LISFLOOD tss file without header (table)
            numserie = len(rec.split())
            outlets_id = []
            for i in range(1, numserie):
                outlets_id.append(i)  #Lisflood progressive ID code for output points
            # read tss data
            # tssdata = pd.read_table(tssfilename, delim_whitespace=True, header=None, names=outlets_id, index_col=0)
    fp.close()
    return outlets_id


def nanCheckMap(data, filename, name):
    """Checks for numpy.nan on simulated pixels: if any is found, a warning is raised"""
    is_nan = np.isnan(compressArray(data) if isinstance(data, pcraster._pcraster.Field) else data)
    if is_nan.any():
        warnings.warn(LisfloodWarning("Warning: {} of {} land values of {} (binding: '{}') are NaN".format(is_nan.sum(), is_nan.size, filename, name)))



#cython: language_level=2, boundscheck=False
"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

Collection of functions called by the Python module kinematic_wave_parallel.py. These functions are
implemented in Cython to achieve compiled-code performance when executing loops over numpy arrays.
To compile this Cython module to enable OpenMP multithreading, execute:
$ python compile_kinematic_wave_parallel_tools.py build_ext --inplace
Important: the module has to be compiled on the machine where the model is run - the resulting binary is not portable.
"""

__author__ = "Emiliano Gelati"
__contact__ = "emiliano.gelati@ec.europa.eu"
__version__ = "1.0"
__date__ = "2016/06/01"

import numpy as np
from cython import boundscheck, wraparound, cdivision
from cython.parallel import prange
from libc.math cimport fabs, fmax, fmin

cdef:
    double NEWTON_TOL = 1e-12 # Max allowed error in iterative solution
    long MAX_ITERS = 3000 # Max number of iterations
    double MIN_DISCHARGE = 1e-30 # Minimum allowed discharge value to avoid numerical instability


# -------------------------------------------------------------------------------------------------
# ROUTING FUNCTIONS
# -------------------------------------------------------------------------------------------------
@boundscheck(False)
@wraparound(False)
def kinematicRouting(double[::1] discharge, double[::1] lateral_inflow, double[::1] constant, long[:,::1] upstream_lookup,\
                     long[::1] num_upstream_pixels, long[::1] ordered_pixels, long[:,::1] start_stop, double beta, double inv_beta,\
                     double b_minus_1, double[::1] a_dx_div_dt, double[::1] b_a_dx_div_dt, long num_threads):
    """"""
    cdef:
        long order, index, first, last, num_orders = start_stop.shape[0]
    # Iterate through routing orders (sets of pixels for which the kinemativc wave can be solved independently and thus in parallel)
    for order in range(num_orders):
        first = start_stop[order,0]
        last = start_stop[order,1]
        for index in prange(first, last, nogil=True, schedule="dynamic", num_threads=num_threads):
            solve1Pixel(ordered_pixels[index], discharge, lateral_inflow, constant, upstream_lookup,\
                        num_upstream_pixels, a_dx_div_dt, b_a_dx_div_dt, beta, inv_beta, b_minus_1)

@boundscheck(False)
@wraparound(False)
@cdivision(True)
cdef void solve1Pixel(long pix, double[::1] discharge, double[::1] lateral_inflow, double[::1] constant,\
                      long[:,::1] upstream_lookup, long[::1] num_upstream_pixels, double[::1] a_dx_div_dt,\
                      double[::1] b_a_dx_div_dt, double beta, double inv_beta, double b_minus_1) nogil:
    """"""
    cdef:
        long ups_ix, count = 0
        double error, const_plus_ups_infl, a_cpui_pow_b_m_1, secant_bound, other_bound, previous_estimate = -1., upstream_inflow = 0.
    # Inflow from upstream pixels
    for ups_ix in range(num_upstream_pixels[pix]):
        upstream_inflow += discharge[upstream_lookup[pix,ups_ix]]
    const_plus_ups_infl = upstream_inflow + constant[pix] # upstream_inflow + alpha*dx/dt*Qold**beta + dx*specific_lateral_inflow
    # If old discharge, upstream inflow and lateral inflow are below accuracy: set discharge to 0 and exit
    if const_plus_ups_infl <= NEWTON_TOL:
        discharge[pix] = 0
        return
    # Initial discharge guess using analytically derived boundary values
    a_cpui_pow_b_m_1 = b_a_dx_div_dt[pix] * const_plus_ups_infl**b_minus_1
    if a_cpui_pow_b_m_1 <= 1:
        secant_bound = const_plus_ups_infl / (1 + a_cpui_pow_b_m_1)
    else:
        secant_bound = const_plus_ups_infl / (1 + a_cpui_pow_b_m_1**inv_beta)
    other_bound = ((const_plus_ups_infl - secant_bound) / a_dx_div_dt[pix])**inv_beta
    discharge[pix] = (secant_bound + other_bound) / 2
    error = closureError(discharge[pix], const_plus_ups_infl, a_dx_div_dt[pix], beta)
    # Iterations
    while fabs(error) > NEWTON_TOL and discharge[pix] != previous_estimate and count < MAX_ITERS: # is previous_estimate useful?
        previous_estimate = discharge[pix]
        discharge[pix] -= error / (1 + b_a_dx_div_dt[pix] * discharge[pix]**b_minus_1)
        discharge[pix] = fmax(discharge[pix], NEWTON_TOL)
        error = closureError(discharge[pix], const_plus_ups_infl, a_dx_div_dt[pix], beta)
        count += 1
    # If iterations converge to NEWTON_TOL, set value to 0
    if discharge[pix] == NEWTON_TOL:
        discharge[pix] = 0

@boundscheck(False)
@wraparound(False)
cdef double closureError(double discharge, double upper_bound, double a_dx_div_dt, double beta) nogil:
    """"""
    return discharge + a_dx_div_dt * discharge**beta - upper_bound



# -------------------------------------------------------------------------------------------------
# FLOW DIRECTION MATRIX PRE-PROCESSING FUNCTIONS
# -------------------------------------------------------------------------------------------------
def updateUpstreamRouted(long[::1] pixels_routed, long[::1] downstream_lookup,\
                         long[:,::1] upstream_lookup, unsigned char[:,::1] upstream_routed):
    """Called by kinematic_wave_parallel.orderKinematicRouting to implement greedy pixel ordering starting from headwaters."""
    cdef:
        long downs_pix, pix, ups_index
    for pix in pixels_routed:
        downs_pix = downstream_lookup[pix]
        if downs_pix != -1:
            ups_index = 0
            while upstream_lookup[downs_pix,ups_index] != pix:
                ups_index += 1
            upstream_routed[downs_pix,ups_index] = True

def upDownLookups(long[:,::1] flow_d8, unsigned char[:,::1] land_mask, long[:,::1] land_points, long num_pixs, long[:,::1] ix_adds):
    '''Called by catchment.streamLookups'''
    cdef:
        long[::1] downstream_lookup = -np.ones(num_pixs, int)
        long[:,::1] upstream_lookup = -np.ones((num_pixs, 8), int)
        unsigned int[::1] ups_count = np.zeros(num_pixs, np.uintc)
        int src_row, src_col, dst_row, dst_col
        unsigned int ups_p, downs_p, num_row = flow_d8.shape[0], num_col = flow_d8.shape[1]
    for src_row in range(num_row):
        for src_col in range(num_col):
            if flow_d8[src_row,src_col] < 8:
                dst_row = src_row + ix_adds[flow_d8[src_row,src_col],0]
                dst_col = src_col + ix_adds[flow_d8[src_row,src_col],1]
                if dst_row != -1 and dst_col != -1 and dst_row != num_row and dst_col != num_col and land_mask[dst_row,dst_col]:
                    ups_p = land_points[src_row,src_col]
                    downs_p = land_points[dst_row,dst_col]
                    downstream_lookup[ups_p] = downs_p
                    upstream_lookup[downs_p,ups_count[downs_p]] = ups_p
                    ups_count[downs_p] += 1
    return np.asarray(downstream_lookup), np.asarray(upstream_lookup)



# -------------------------------------------------------------------------------------------------
# MISCELLANOUS TOOLS
# -------------------------------------------------------------------------------------------------
def immediateUpstreamInflow(double[::1] discharge, long[:,::1] upstream_lookup, long[::1] num_upstream_pixels):
    """"""
    cdef:
        long pix, ups_ix, num_pixels = discharge.size
        double [::1] inflow = np.zeros(num_pixels)
    for pix in range(num_pixels):
        for ups_ix in range(num_upstream_pixels[pix]):
            inflow[pix] += discharge[upstream_lookup[pix,ups_ix]]
    return np.asarray(inflow)



"""

Copyright 2019-2021 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""
from __future__ import print_function, absolute_import
from nine import range

import uuid
import sys
import datetime
import gc

from pcraster.framework import DynamicModel
import numpy as np

from .global_modules.settings import CDFFlags, LisSettings, MaskInfo

class LisfloodModel_dyn(DynamicModel):

    # =========== DYNAMIC ====================================================

    def dynamic(self):
        """ Dynamic part of LISFLOOD
            calls the dynamic part of the hydrological modules
        """
        settings = LisSettings.instance()
        option = settings.options
        flags = settings.flags
        # date corresponding to the model time step (yyyy-mm-dd hh:mm:ss)
        self.CalendarDate = self.CalendarDayStart + datetime.timedelta(days=(self.currentTimeStep()-1) * self.DtDay)
        # day of the year corresponding to the model time step
        self.CalendarDay = int(self.CalendarDate.strftime("%j"))
        # correct method to calculate the day of the year

        # model time step
        i = self.currentTimeStep()
        if i == 1:
            # flag for netcdf output for all, steps and end
            _ = CDFFlags(uuid.uuid4())  # init CDF flags

        self.TimeSinceStart = self.currentTimeStep() - self.firstTimeStep() + 1
        if flags['loud']:
            print("%-6i %10s" % (self.currentTimeStep(), self.CalendarDate.strftime("%d/%m/%Y %H:%M")))
        else:
            if not flags['checkfiles']:
                if flags['quiet'] and not flags['veryquiet']:
                    sys.stdout.write(".")
                if not flags['quiet'] and not flags['veryquiet']:
                    # Print step number and date to console
                    sys.stdout.write("\r%d" % i), sys.stdout.write("%s" % " - "+self.CalendarDate.strftime("%d/%m/%Y %H:%M"))
                    sys.stdout.flush()
        if i == self.nrTimeSteps():
            # last timestep. Send a new line to the terminal for polishness
            if not flags['veryquiet']:
                sys.stdout.write('\n')
            sys.stdout.flush()

        # ************************************************************
        """ up to here it was fun, now the real stuff starts
        """
        # readmeteo.py
        self.readmeteo_module.dynamic()                 
            
        if flags['checkfiles']:
            return  # if check than finish here

        """ Here it starts with hydrological modules:
        """

        # ***** READ land use fraction maps***************************
        self.landusechange_module.dynamic()

        # ***** READ LEAF AREA INDEX DATA ****************************
        self.leafarea_module.dynamic()

        # ***** READ variable water fraction ****************************
        self.evapowater_module.dynamic_init()

        # ***** READ INFLOW HYDROGRAPHS (OPTIONAL)****************
        self.inflow_module.dynamic()

        # ***** RAIN AND SNOW *****************************************
        self.snow_module.dynamic()

        # ***** FROST INDEX IN SOIL **********************************
        self.frost_module.dynamic()

        # ***** EPIC AGRICULTURE MODEL - 1ST PART: CROP STATE AND ENVIRONMENT *******************
        if option["cropsEPIC"]:
            self.crop_module.dynamic_state()

        # *************************************************************************************
        # **** Loop over vegetation fractions: 1. processes depending directly on the canopy
        # *************************************************************************************
        self.soilloop_module.dynamic_canopy()

        # ***** EPIC AGRICULTURE MODEL - 2ND PART: CROP GROWTH AND LIMITNG FACTORS *************
        if option["cropsEPIC"]:
            self.crop_module.dynamic_growth()
  
        # **************************************************************************************
        # **** Loop over vegetation fractions: 2. internal soil processes
        # **************************************************************************************
        self.soilloop_module.dynamic_soil()

        # -------------------------------------------------------------------
        # -------------------------------------------------------------------

        # ***** ACTUAL EVAPORATION FROM OPEN WATER AND SEALED SOIL ***
        self.opensealed_module.dynamic()

        # *********  WATER USE   *************************
        self.riceirrigation_module.dynamic()
        self.waterabstraction_module.dynamic()
         
        ## repeating lines below to create a variable with the correct structure ##
        maskinfo = MaskInfo.instance()
        def splitlanduse(array1, array2=None, array3=None):
            """ splits maps into the 3 different land use types - other , forest, irrigation
            """
            if array2 is None:
                array2 = array1
            if array3 is None:
                array3 = array1
            return [array1, array2, array3]
         
        # ***** Calculation per Pixel ********************************
        self.soil_module.dynamic_perpixel()
 
        self.groundwater_module.dynamic()
 
        # ************************************************************
        # ***** STOP if no routing is required    ********************
        # ************************************************************
        if option['InitLisfloodwithoutSplit']:
            # InitLisfloodwithoutSplit
            # Very fast InitLisflood
            # it is only to compute Lzavin.map and skip completely the routing component
            self.output_module.dynamic()  # only lzavin
            return

        # *********  EVAPORATION FROM OPEN WATER *************
        self.evapowater_module.dynamic()
 
        # ***** ROUTING SURFACE RUNOFF TO CHANNEL ********************
        self.surface_routing_module.dynamic()
 
        # ***** POLDER INIT **********************************
        self.polder_module.dynamic_init()

        # ***** INLETS INIT **********************************
        self.inflow_module.dynamic_init()
 
        # ************************************************************
        # ***** LOOP ROUTING SUB TIME STEP   *************************
        # ************************************************************
        maskinfo = MaskInfo.instance()
        self.sumDisDay = maskinfo.in_zero()
        # sums up discharge of the sub steps
        for NoRoutingExecuted in range(self.NoRoutSteps):
            self.routing_module.dynamic(NoRoutingExecuted)
            #   routing sub steps
 
        # ----------------------------------------------------------------------

        if option['inflow']:
            self.QInM3Old = self.QInM3
            # to calculate the parts of inflow for every routing timestep
            # for the next timestep the old inflow is preserved
            self.sumInWB = self.QinADDEDM3

        # if option['simulatePolders']:
        # ChannelToPolderM3=ChannelToPolderM3Old;

        if option['InitLisflood'] or (not(option['SplitRouting'])):
            # kinematic routing
            self.ChanM3 = self.ChanM3Kin.copy()
            # Total channel storage [cu m], equal to ChanM3Kin
        else:
            # split routing
            self.ChanM3 = self.ChanM3Kin + self.Chan2M3Kin - self.Chan2M3Start

        # Total channel storage [cu m], equal to ChanM3Kin
        # sum of both lines
        # CrossSection2Area = pcraster.max(scalar(0.0), (self.Chan2M3Kin - self.Chan2M3Start) / self.ChanLength)

        self.TotalCrossSectionArea = self.ChanM3 * self.InvChanLength

        self.sumDis += self.sumDisDay
        self.ChanQAvg = self.sumDisDay/self.NoRoutSteps

        # Total volume of water in channel per inv channel length
        # New cross section area (kinematic wave)
        # This is the value after the kinematic wave, so we use ChanM3Kin here
        # (NOT ChanQKin, which is average discharge over whole step, we need state at the end of all iterations!)

        # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        if not(option['dynamicWave']):
            # Dummy code if dynamic wave is not used, in which case the total cross-section
            # area equals TotalCrossSectionAreaKin, ChanM3 equals ChanM3Kin and
            # ChanQ equals ChanQKin
            WaterLevelDyn = -9999
            # Set water level dynamic wave to dummy value (needed

        if option['InitLisflood'] or option['repAverageDis']:
            self.CumQ += self.ChanQ
            self.avgdis = self.CumQ/self.TimeSinceStart
            # to calculate average discharge

        self.DischargeM3Out += np.where(self.AtLastPointC ,self.ChanQ * self.DtSec,0)
           # Cumulative outflow out of map

        # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        # Calculate water level
        self.waterlevel_module.dynamic()

        # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        # ************************************************************
        # *******  Calculate CUMULATIVE MASS BALANCE ERROR  **********
        # ************************************************************
        self.waterbalance_module.dynamic()
        if option['indicator']:
           self.indicatorcalc_module.dynamic()

        # ************************************************************
        # ***** WRITING RESULTS: TIME SERIES AND MAPS ****************
        # ************************************************************
        self.output_module.dynamic()

        # debug 
        # Print value of variables after computation (from state files)
        if flags['debug']:
            nomefile = 'Debug_out_'+str(self.currentStep)+'.txt'
            ftemp1 = open(nomefile, 'w+')
            nelements = len(self.ChanM3)
            for i in range(0,nelements-1):
                if  hasattr(self,'CrossSection2Area'):
                    print(i, self.TotalCrossSectionArea[i], self.CrossSection2Area[i], self.ChanM3[i], \
                    self.Chan2M3Kin[i], file=ftemp1)
                else:
                    print(i, self.TotalCrossSectionArea[i], self.ChanM3[i], file=ftemp1)
            ftemp1.close()

        ### Report states if EnKF is used and filter moment
        self.stateVar_module.dynamic()
        if option['wateruse'] and option['indicator'] and self.monthend: 
            self.indicatorcalc_module.dynamic_setzero()
        # setting monthly and yearly dindicator to zero at the end of the month (year)



import os
current_dir = os.path.dirname(os.path.abspath(__file__))

version_file = os.path.join(current_dir, '../../VERSION')
if not os.path.exists(version_file):
    version_file = os.path.join(current_dir, '../../../../VERSION')

with open(version_file, 'r') as f:
    version = f.read().strip()

__version__ = version
__authors__ = "Ad de Roo, Emiliano Gelati, Peter Burek, Johan van der Knijff, Niko Wanders"
__date__ = "09/06/2022"
__copyright__ = "Copyright 2019-2022, European Commission - Joint Research Centre"
__maintainer__ = "Stefania Grimaldi, Cinzia Mazzetti, Carlo Russo, Damien Decremer, Corentin Carton De Wiart, Valerio Lorini, Ad de Roo"
__status__ = "Operation"



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""
from __future__ import print_function, absolute_import

from ..global_modules.settings import LisSettings
from ..global_modules.add1 import readmapsparse
from ..global_modules.netcdf import xarray_reader


class readmeteo(object):

    """
     # ************************************************************
     # ***** READ METEOROLOGICAL DATA              ****************
     # ************************************************************
    """

    def __init__(self, readmeteo_variable):
        self.var = readmeteo_variable
        settings = LisSettings.instance()
        option = settings.options
        # initialise xarray readers
        if option['readNetcdfStack']:
            self.forcings = {}
            for data in ['PrecipitationMaps', 'TavgMaps', 'ET0Maps', 'E0Maps']:
                self.forcings[data] = xarray_reader(data)

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
    def dynamic(self):
        """ dynamic part of the readmeteo module
            read meteo input maps
        """
        settings = LisSettings.instance()
        option = settings.options
        binding = settings.binding

        # ************************************************************
        # ***** READ METEOROLOGICAL DATA *****************************
        # ************************************************************
        if option['readNetcdfStack']:

            step = self.var.currentTimeStep() - self.var.firstTimeStep()
            # date = date_from_step(binding, self.var.currentTimeStep())

            # Read from NetCDF stack files
            self.var.Precipitation = self.forcings['PrecipitationMaps'][step] * self.var.DtDay * self.var.PrScaling
            self.var.Tavg = self.forcings['TavgMaps'][step]
            self.var.ETRef = self.forcings['ET0Maps'][step] * self.var.DtDay * self.var.CalEvaporation
            self.var.EWRef =self.forcings['E0Maps'][step] * self.var.DtDay * self.var.CalEvaporation

        else:
            # Read from stack of maps in Pcraster format
            self.var.Precipitation = readmapsparse(binding['PrecipitationMaps'], self.var.currentTimeStep(), self.var.Precipitation) * self.var.DtDay * self.var.PrScaling
            # precipitation (conversion to [mm] per time step)
            self.var.Tavg = readmapsparse(binding['TavgMaps'], self.var.currentTimeStep(), self.var.Tavg)
            # average DAILY temperature (even if you are running the model on say an hourly time step) [degrees C]
            self.var.ETRef = readmapsparse(binding['ET0Maps'], self.var.currentTimeStep(), self.var.ETRef) * self.var.DtDay * self.var.CalEvaporation
            # daily reference evapotranspiration (conversion to [mm] per time step)
            # potential evaporation rate from a bare soil surface (conversion to [mm] per time step)
            self.var.EWRef = readmapsparse(binding['E0Maps'], self.var.currentTimeStep(), self.var.EWRef) * self.var.DtDay * self.var.CalEvaporation
            # potential evaporation rate from water surface (conversion to [mm] per time step)

        self.var.ESRef = (self.var.EWRef + self.var.ETRef)/2

        if option['TemperatureInKelvin']:
            self.var.Tavg -= 273.15



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""
from __future__ import print_function, absolute_import, unicode_literals

from nine import range

from platform import system as operating_system
import time as xtime
from netCDF4 import Dataset

try:
    from netCDF4 import netcdftime
except ImportError:
    import cftime as netcdftime  # newer versions of netCDF4 don't include netcdftime

from pcraster.framework import *  # TODO fix import star
import numpy as np

from .decorators import counted
from .settings import datetoint, LisSettings, MaskInfo
from .errors import LisfloodError, LisfloodFileError


MAX_READ_TRIALS = 100 # max number of trial allowed re-read an input file: to avoid crashes due to temporary network interruptions
READ_PAUSE = 0.1      # pause (seconds) between each re-read trial over the network
try:
    NC_DATE_TYPE = netcdftime._netcdftime.datetime
except:
    NC_DATE_TYPE = netcdftime.datetime  # work-around for older versions of the netCDF4 library (< 1.2.7 ?)


project_dir = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '../..'))


@counted
def checkmap(name, value, map_to_check, flagmap, find):
    """ Check if maps fit to the mask mamap_to_checkp
    
    :param name: key in Settings.xml containing path and name of the map to be checked as string (map name)
    :param value: name and path of the map to be checked as string
    :param map_to_check: map to be used for checking in pcraster format
    :param flagmap: flag for maps
    :param find: 
    :return: 
    """
    s = [name, value]
    maskinfo = MaskInfo.instance()
    maskmap = maskinfo.maskmap
    if flagmap:
        amap = scalar(defined(maskmap))
        try:
            smap = scalar(defined(map_to_check))
        except:
            msg = "Map: " + name + " in " + value + " does not fit"
            if name == "LZAvInflowMap":
                msg +="\nMaybe run initial run first"
            raise LisfloodError(msg)

        mvmap = maptotal(smap)
        mv = cellvalue(mvmap, 1, 1)[0]
        s.append(mv)

        less = maptotal(ifthenelse(defined(maskmap), amap - smap, scalar(0)))
        s.append(cellvalue(less, 1, 1)[0])
        less = mapminimum(scalar(map_to_check))
        s.append(cellvalue(less, 1, 1)[0])
        less = maptotal(scalar(map_to_check))
        if mv > 0:
            s.append(cellvalue(less, 1, 1)[0] / mv)
        else:
            s.append('0')
        less = mapmaximum(scalar(map_to_check))
        s.append(cellvalue(less, 1, 1)[0])
        if find > 0:
            if find == 2:
                s.append('last_Map_used')
            else:
                s.append('')

    else:
        s.append(0)
        s.append(0)
        s.append(float(map_to_check))
        s.append(float(map_to_check))
        s.append(float(map_to_check))

    # print check results
    if checkmap.called == 1:
        print("%-25s%-40s%11s%11s%11s%11s%11s" %("Name","File/Value","nonMV","MV","min","mean","max"))
    print("%-25s%-40s%11i%11i%11.2f%11.2f%11.2f" %(s[0],s[1][-39:],s[2],s[3],s[4],s[5],s[6]))
    return


class DynamicFramework(DynamicFramework):
    """
    Framework class for dynamic models.
    `userModel`
    Instance that models the :ref:`Dynamic Model Concept <dynamicModelConcept>`.
    `lastTimeStep`   Last timestep to run.
    `firstTimestep`  Sets the starting timestep of the model (optional, default is 1).
    Updated by zusatz.py
    """

    def run(self):
        """
        Run the dynamic user model.

        .. todo::
        This method depends on the filter frameworks concept. Shouldn't its run
        method call _runSuspend()?
        """
        settings = LisSettings.instance()
        binding = settings.binding
        self._atStartOfScript()
        if hasattr(self._userModel(), "resume"):
            # if self._userModel().firstTimeStep() == 1:
            # replaced this because starting date is not always the 1
            if self._userModel().firstTimeStep() == datetoint(binding['StepStart'])[0]:
                self._runInitial()
            else:
                self._runResume()
        else:
            self._runInitial()

        self._runDynamic()

        # Only execute this section while running filter frameworks.
        if hasattr(self._userModel(), "suspend") and hasattr(self._userModel(), "filterPeriod"):
            self._runSuspend()

        return 0

    """Adjusting the def _atStartOfTimeStep defined in DynamicFramework
       for a real quiet output
    """
    rquiet = False
    rtrace = False

    def _atStartOfTimeStep(self, step):
        self._userModel()._setInTimeStep(True)
        if not self.rquiet:
            if not self.rtrace:
                msg = u"#"

            else:
                msg = u"%s<time step=\"%s\">\n" % (self._indentLevel(), step)
            sys.stdout.write(msg)
            sys.stdout.flush()


class EnsKalmanFilterFramework(EnsKalmanFilterFramework):
    """
    Framework for particle filter runs
    Updated by zusatz
    """
    def _startEndOfPeriod(self, currentPeriod, lastPeriod):
        # determine start end end timestep of current period
        if currentPeriod == 0:
          #startTimestep = 1  # use first timestep instead of hardcode 1
          startTimestep = self._userModel().firstTimeStep()

          endTimestep = self._userModel()._d_filterTimesteps[currentPeriod]
        elif currentPeriod == lastPeriod:
          startTimestep = self._userModel()._d_filterTimesteps[currentPeriod -1] + 1
          endTimestep = self._d_totalTimesteps
        else:
          startTimestep = self._userModel()._d_filterTimesteps[currentPeriod - 1] + 1
          endTimestep = self._userModel()._d_filterTimesteps[currentPeriod]

        assert startTimestep <= endTimestep
        return startTimestep, endTimestep


class TimeoutputTimeseries(TimeoutputTimeseries):
    """
    Class to create pcrcalc timeoutput style timeseries
    Updated py zusatz.py
    """
    def _writeFileHeader(self, outputFilename):
        """
        writes header part of tss file
        """
        outputFile = open(outputFilename, "w")
        # header
        from .settings import LisSettings
        settings = LisSettings.instance()
        outputFile.write("timeseries {} settingsfile: {} date: {}\n".format(self._spatialDatatype.lower(),
                                                                            settings.settings_path,
                                                                            xtime.ctime(xtime.time())))
        # write number of outlets points +1
        outputFile.write(str(self._ncodesId + 1) + "\n")
        outputFile.write("timestep\n")

        for colId in range(0, self._ncodesId):
            outputFile.write(str(self._codesId[colId]) + "\n")
        outputFile.close()

    def _configureOutputFilename(self, filename):
        """
        generates filename
        appends timeseries file extension if necessary
        prepends sample directory if used in stochastic
        """

        # test if suffix or path is given
        head, tail = os.path.split(filename)

        if not re.search("\.tss", tail):
            # content,sep,comment = filename.partition("-")
            # filename = content + "Tss" + sep + comment + ".tss"
            filename = tail + ".tss"

        # for stochastic add sample directory
        if hasattr(self._userModel, "nrSamples"):
            try:
                filename = os.path.join(str(self._userModel.currentSampleNumber()), filename)
            except:
                pass

        return filename

    def _writeTssFile(self):
        """
        writing timeseries to disk
        """
        #
        option = LisSettings.instance().options
        outputFilename = self._configureOutputFilename(self._outputFilename)
        if option['EnKF']:
            if not os.path.exists(outputFilename):
                if self._writeHeader:
                   self._writeFileHeader(outputFilename)
                   outputFile = open(outputFilename, "a")
                else:
                   outputFile = open(outputFilename, "w")
            else:
                 outputFile = open(outputFilename, "a")
        else:
            if self._writeHeader:
                  self._writeFileHeader(outputFilename)
                  outputFile = open(outputFilename, "a")
            else:
                  outputFile = open(outputFilename, "w")

        assert outputFile

        start = self._userModel.firstTimeStep()
        end = self._userModel.nrTimeSteps() + 1

        for timestep in range(start, end):
            row = ""
            row += " %8g" % timestep
            if self._spatialIdGiven:
                for cellId in range(0, self._ncodesId):
                    value = self._sampleValues[timestep - start][cellId]
                    if isinstance(value, Decimal):
                        row += "           1e31"
                    else:
                        row += " %14g" % (value)
                row += "\n"
            else:
                value = self._sampleValues[timestep - start]
                if isinstance(value, Decimal):
                    row += "           1e31"
                else:
                    row += " %14g" % (value)
                row += "\n"

            outputFile.write(row)
        outputFile.close()

    def __init__(self, tssFilename, model, idMap=None, noHeader=False):
        """

        """

        if not isinstance(tssFilename, str):
            raise ValueError("timeseries output filename must be of type string. Found {} of type {}".format(tssFilename, type(tssFilename)))
        settings = LisSettings.instance()
        binding = settings.binding
        self._outputFilename = tssFilename
        self._maxId = 1
        self._ncodesId = 1
        self._spatialId = None
        self._spatialDatatype = None
        self._spatialIdGiven = False
        self._userModel = model
        self._writeHeader = not noHeader
        # array to store the timestep values
        self._sampleValues = None

        _idMap = False
        if isinstance(idMap, str) or isinstance(idMap, pcraster._pcraster.Field):
            _idMap = True

        # if header reserve rows from 1 to endstep
        # if noheader only from startstep - endstep

        if noHeader:
            nrRows = datetoint(binding['StepEnd'])[0] - datetoint(binding['StepStart'])[0] - self._userModel.firstTimeStep() + 2
        else:
            nrRows = datetoint(binding['StepEnd'])[0] - datetoint(binding['StepStart'])[0] - self._userModel.firstTimeStep() + 2

        if _idMap:
            self._spatialId = idMap
            if isinstance(idMap, str):
                self._spatialId = iterReadPCRasterMap(idMap)

            _allowdDataTypes = [
                pcraster.Nominal, pcraster.Ordinal, pcraster.Boolean]
            if self._spatialId.dataType() not in _allowdDataTypes:
                #raise Exception(
                #    "idMap must be of type Nominal, Ordinal or Boolean")
                # changed into creating a nominal map instead of bailing out
                self._spatialId = pcraster.nominal(self._spatialId)

            if self._spatialId.isSpatial():
                self._maxId, valid = pcraster.cellvalue(
                    pcraster.mapmaximum(pcraster.ordinal(self._spatialId)), 1)
                # convert to numpy array
                outletsmapnp = pcr2numpy(self._spatialId,np.nan)
                # get outlets codes from outlets map
                codesId = numpy.unique(outletsmapnp)
                # drop negative values (= missing data in pcraster map)
                codesId = codesId[codesId > 0]
                # get number of outlets points
                self._ncodesId = len(codesId)
                # prepare array to store outlets codes
                self._codesId = [-9999 for i in range(self._ncodesId)]

            else:
                self._maxId = 1
                self._codesId = [1]
                self._ncodesId = len(self._codesId)

            # cell indices of the sample locations

            # #self._sampleAddresses = []
            # for cellId in range(1, self._maxId + 1):
            # self._sampleAddresses.append(self._getIndex(cellId))
            # prepare array to store outlets points raster numbers
            self._sampleAddresses = [-9999 for i in range(self._ncodesId)]
            # init with the left/top cell - could also be 0 but then you have to catch it in
            # the sample routine and put an exeption in
            # number of cells in map
            nrCells = pcraster.clone().nrRows() * pcraster.clone().nrCols()
            for cell in range(1, nrCells + 1):
                if (pcraster.cellvalue(self._spatialId, cell)[1]):
                    # get point code from outlets map for pixel cell
                    outlet_code = pcraster.cellvalue(self._spatialId, cell)[0]
                    # get index of the point code in the sorted list of outlets codes
                    outlet_idx = np.where(codesId == outlet_code)[0][0]
                    # store point code
                    self._codesId[outlet_idx] = outlet_code
                    # store outlets location (cell)
                    self._sampleAddresses[outlet_idx] = cell

            self._spatialIdGiven = True

            nrCols = self._ncodesId
            self._sampleValues = [
                [Decimal("NaN")] * nrCols for _ in [0] * nrRows]
        else:
            self._sampleValues = [[Decimal("NaN")] * 1 for _ in [0] * nrRows]

    def firstout(self,expression):
        """
        returns the first cell as output value
        """
        try:
            cellIndex = self._sampleAddresses[0]
            tmp = pcraster.areaaverage(pcraster.spatial(expression), pcraster.spatial(self._spatialId))
            value, valid = pcraster.cellvalue(tmp, cellIndex)
            if not valid:
               value = Decimal("NaN")
        except:
            value = Decimal("NaN")
        return value


#####################################################################################################################
# TOOLS TO OPEN/READ INPUT FILES ITERATIVELY, IN CASE OF NETWORK TEMPORARILY DOWN
#####################################################################################################################

def iterOpenNetcdf(file_path, error_msg, mode, **kwargs):
    """Wrapper around netCDF4.Dataset function exploiting the iterAccess class to access file_path according to the specified mode"""
    access_function = lambda path: Dataset(path, mode, **kwargs)
    return remoteInputAccess(access_function, file_path, error_msg)


def iterReadPCRasterMap(file_path, error_msg=""):
    """Wrapper around pcraster.readmap function exploiting the iterAccess class to open file_path."""
    return remoteInputAccess(pcraster.readmap, file_path, error_msg)


def iterSetClonePCR(file_path, error_msg=""):
    """Wrapper around pcraster.setclone function exploiting the iterAccess class to access file_path."""
    return remoteInputAccess(pcraster.setclone, file_path, error_msg)


def remoteInputAccess(function, file_path, error_msg):
    """
    Wrapper around the provided file access function.
    It allows re-trying the open/read operation if network is temporarily down.
    Arguments:
        function: function to be called to read/open the file.
        file_path: path of the file to be read/open.
    """
    num_trials = 1
    bad_sep = "/" if operating_system() == "Windows" else "\\"
    file_path = file_path.replace(bad_sep, os.path.sep)
    root = os.path.sep.join(file_path.split(os.path.sep)[:4])

    while num_trials <= MAX_READ_TRIALS:
        try:
            obj = function(str(file_path))
            if num_trials > 1:
                print("File {0} succesfully accessed after {1} attempts".format(file_path, num_trials))
            num_trials = MAX_READ_TRIALS + 1
        except IOError:
            if os.path.exists(root) and not os.path.exists(file_path):
                raise LisfloodFileError(file_path, error_msg)
            elif num_trials == MAX_READ_TRIALS:
                raise Exception("Cannot access file {0}!\nNetwork down for too long OR bad root directory {1}!".format(file_path, root))
            else:
                num_trials += 1
                print("Trying to access file {0}: attempt n. {1}".format(file_path, num_trials))
                xtime.sleep(READ_PAUSE)
    return obj
